# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import warnings
from typing import Callable, Sequence

import numpy as np
import torch

from monai.config.type_definitions import DtypeLike, NdarrayOrTensor
from monai.transforms import (
    Activations,
    AsDiscrete,
    BoundingRect,
    FillHoles,
    GaussianSmooth,
    RemoveSmallObjects,
    SobelGradients,
)
from monai.transforms.transform import Transform
from monai.transforms.utils_pytorch_numpy_unification import max, maximum, min, sum, unique
from monai.utils import TransformBackends, convert_to_numpy, optional_import
from monai.utils.misc import ensure_tuple_rep
from monai.utils.type_conversion import convert_to_dst_type, convert_to_tensor

label, _ = optional_import("scipy.ndimage.measurements", name="label")
disk, _ = optional_import("skimage.morphology", name="disk")
opening, _ = optional_import("skimage.morphology", name="opening")
watershed, _ = optional_import("skimage.segmentation", name="watershed")
find_contours, _ = optional_import("skimage.measure", name="find_contours")
centroid, _ = optional_import("skimage.measure", name="centroid")

__all__ = [
    "Watershed",
    "GenerateWatershedMask",
    "GenerateInstanceBorder",
    "GenerateDistanceMap",
    "GenerateWatershedMarkers",
    "GenerateSuccinctContour",
    "GenerateInstanceContour",
    "GenerateInstanceCentroid",
    "GenerateInstanceType",
    "HoVerNetInstanceMapPostProcessing",
    "HoVerNetNuclearTypePostProcessing",
]


class Watershed(Transform):
    """
    Use `skimage.segmentation.watershed` to get instance segmentation results from images.
    See: https://scikit-image.org/docs/stable/api/skimage.segmentation.html#skimage.segmentation.watershed.

    Args:
        connectivity: an array with the same number of dimensions as image whose non-zero elements indicate
            neighbors for connection. Following the scipy convention, default is a one-connected array of
            the dimension of the image.
        dtype: target data content type to convert, default is np.int64.

    """

    backend = [TransformBackends.NUMPY]

    def __init__(self, connectivity: int | None = 1, dtype: DtypeLike = np.int64) -> None:
        self.connectivity = connectivity
        self.dtype = dtype

    def __call__(
        self, image: NdarrayOrTensor, mask: NdarrayOrTensor | None = None, markers: NdarrayOrTensor | None = None
    ) -> NdarrayOrTensor:
        """
        Args:
            image: image where the lowest value points are labeled first. Shape must be [1, H, W, [D]].
            mask: optional, the same shape as image. Only points at which mask == True will be labeled.
                If None (no mask given), it is a volume of all 1s.
            markers: optional, the same shape as image. The desired number of markers, or an array marking
                the basins with the values to be assigned in the label matrix. Zero means not a marker.
                If None (no markers given), the local minima of the image are used as markers.
        """

        image = convert_to_numpy(image)
        markers = convert_to_numpy(markers)
        mask = convert_to_numpy(mask)

        instance_seg = watershed(image, markers=markers, mask=mask, connectivity=self.connectivity)

        return convert_to_dst_type(instance_seg, image, dtype=self.dtype)[0]


class GenerateWatershedMask(Transform):
    """
    generate mask used in `watershed`. Only points at which mask == True will be labeled.

    Args:
        activation: the activation layer to be applied on the input probability map.
            It can be "softmax" or "sigmoid" string, or any callable. Defaults to "softmax".
        threshold: an optional float value to threshold to binarize probability map.
            If not provided, defaults to 0.5 when activation is not "softmax", otherwise None.
        min_object_size: objects smaller than this size (in pixel) are removed. Defaults to 10.
        dtype: target data content type to convert, default is np.uint8.

    """

    backend = [TransformBackends.NUMPY]

    def __init__(
        self,
        activation: str | Callable = "softmax",
        threshold: float | None = None,
        min_object_size: int = 10,
        dtype: DtypeLike = np.uint8,
    ) -> None:
        self.dtype = dtype

        # set activation layer
        use_softmax = False
        use_sigmoid = False
        activation_fn = None
        if isinstance(activation, str):
            if activation.lower() == "softmax":
                use_softmax = True
            elif activation.lower() == "sigmoid":
                use_sigmoid = True
            else:
                raise ValueError(
                    f"The activation should be 'softmax' or 'sigmoid' string, or any callable. '{activation}' was given."
                )
        elif callable(activation):
            activation_fn = activation
        else:
            raise ValueError(f"The activation type should be either str or callable. '{type(activation)}' was given.")
        self.activation = Activations(softmax=use_softmax, sigmoid=use_sigmoid, other=activation_fn)

        # set discretization transform
        if not use_softmax and threshold is None:
            threshold = 0.5
        self.as_discrete = AsDiscrete(threshold=threshold, argmax=use_softmax)

        # set small object removal transform
        self.remove_small_objects = RemoveSmallObjects(min_size=min_object_size) if min_object_size > 0 else None

    def __call__(self, prob_map: NdarrayOrTensor) -> NdarrayOrTensor:
        """
        Args:
            prob_map: probability map of segmentation, shape must be [C, H, W, [D]]
        """

        pred = self.activation(prob_map)
        pred = self.as_discrete(pred)

        pred = convert_to_numpy(pred)

        pred = label(pred)[0]
        if self.remove_small_objects is not None:
            pred = self.remove_small_objects(pred)
        pred[pred > 0] = 1  # type: ignore

        return convert_to_dst_type(pred, prob_map, dtype=self.dtype)[0]


class GenerateInstanceBorder(Transform):
    """
    Generate instance border by hover map. The more parts of the image that cannot be identified as foreground areas,
    the larger the grey scale value. The grey value of the instance's border will be larger.

    Args:
        kernel_size: the size of the Sobel kernel. Defaults to 5.
        dtype: target data type to convert to. Defaults to np.float32.


    Raises:
        ValueError: when the `mask` shape is not [1, H, W].
        ValueError: when the `hover_map` shape is not [2, H, W].

    """

    backend = [TransformBackends.NUMPY]

    def __init__(self, kernel_size: int = 5, dtype: DtypeLike = np.float32) -> None:
        self.dtype = dtype
        self.sobel_gradient = SobelGradients(kernel_size=kernel_size)

    def __call__(self, mask: NdarrayOrTensor, hover_map: NdarrayOrTensor) -> NdarrayOrTensor:  # type: ignore
        """
        Args:
            mask: binary segmentation map, the output of :py:class:`GenerateWatershedMask`.
                Shape must be [1, H, W] or [H, W].
            hover_map:  horizontal and vertical distances of nuclear pixels to their centres of mass. Shape must be [2, H, W].
                The first and second channel represent the horizontal and vertical maps respectively. For more details refer
                to papers: https://arxiv.org/abs/1812.06499.
        """
        if len(hover_map.shape) != 3:
            raise ValueError(f"The hover map should have the shape of [C, H, W], but got {hover_map.shape}.")
        if len(mask.shape) == 3:
            if mask.shape[0] != 1:
                raise ValueError(f"The mask should have only one channel, but got {mask.shape[0]}.")
        elif len(mask.shape) == 2:
            mask = mask[None]
        else:
            raise ValueError(f"The mask should have the shape of [1, H, W] or [H, W], but got {mask.shape}.")
        if hover_map.shape[0] != 2:
            raise ValueError(f"Suppose the hover map only has two channels, but got {hover_map.shape[0]}")

        hover_h = hover_map[0:1, ...]
        hover_v = hover_map[1:2, ...]

        hover_h_min, hover_h_max = min(hover_h), max(hover_h)
        hover_v_min, hover_v_max = min(hover_v), max(hover_v)
        if (hover_h_max - hover_h_min) == 0 or (hover_v_max - hover_v_min) == 0:
            raise ValueError("Not a valid hover map, please check your input")
        hover_h = (hover_h - hover_h_min) / (hover_h_max - hover_h_min)
        hover_v = (hover_v - hover_v_min) / (hover_v_max - hover_v_min)
        sobelh = self.sobel_gradient(hover_h)[1, ...]
        sobelv = self.sobel_gradient(hover_v)[0, ...]
        sobelh_min, sobelh_max = min(sobelh), max(sobelh)
        sobelv_min, sobelv_max = min(sobelv), max(sobelv)
        if (sobelh_max - sobelh_min) == 0 or (sobelv_max - sobelv_min) == 0:
            raise ValueError("Not a valid sobel gradient map")
        sobelh = 1 - (sobelh - sobelh_min) / (sobelh_max - sobelh_min)
        sobelv = 1 - (sobelv - sobelv_min) / (sobelv_max - sobelv_min)

        # combine the h & v values using max
        overall = maximum(sobelh, sobelv)
        overall = overall - (1 - mask)
        overall[overall < 0] = 0

        return convert_to_dst_type(overall, mask, dtype=self.dtype)[0]


class GenerateDistanceMap(Transform):
    """
    Generate distance map.
    In general, the instance map is calculated from the distance to the background.
    Here, we use 1 - "instance border map" to generate the distance map.
    Nuclei values form mountains so invert them to get basins.

    Args:
        smooth_fn: smoothing function for distance map, which can be any callable object.
            If not provided :py:class:`monai.transforms.GaussianSmooth()` is used.
        dtype: target data type to convert to. Defaults to np.float32.
    """

    backend = [TransformBackends.NUMPY]

    def __init__(self, smooth_fn: Callable | None = None, dtype: DtypeLike = np.float32) -> None:
        self.smooth_fn = smooth_fn if smooth_fn is not None else GaussianSmooth()
        self.dtype = dtype

    def __call__(self, mask: NdarrayOrTensor, instance_border: NdarrayOrTensor) -> NdarrayOrTensor:  # type: ignore
        """
        Args:
            mask: binary segmentation map, the output of :py:class:`GenerateWatershedMask`.
                Shape must be [1, H, W] or [H, W].
            instance_border: instance border map, the output of :py:class:`GenerateInstanceBorder`.
                Shape must be [1, H, W].
        """
        if len(mask.shape) == 3:
            if mask.shape[0] != 1:
                raise ValueError(f"The mask should have only one channel, but got {mask.shape[0]}.")
        elif len(mask.shape) == 2:
            mask = mask[None]
        else:
            raise ValueError(f"The mask should have the shape of [1, H, W] or [H, W], but got {mask.shape}.")
        if instance_border.shape[0] != 1 or instance_border.ndim != 3:
            raise ValueError(f"Input instance_border should be with size of [1, H, W], but got {instance_border.shape}")

        distance_map = (1.0 - instance_border) * mask
        distance_map = self.smooth_fn(distance_map)  # type: ignore

        return convert_to_dst_type(-distance_map, mask, dtype=self.dtype)[0]


class GenerateWatershedMarkers(Transform):
    """
    Generate markers to be used in `watershed`. The watershed algorithm treats pixels values as a local topography
    (elevation). The algorithm floods basins from the markers until basins attributed to different markers meet on
    watershed lines. Generally, markers are chosen as local minima of the image, from which basins are flooded.
    Here is the implementation from HoVerNet paper.
    For more details refer to papers: https://arxiv.org/abs/1812.06499.

    Args:
        threshold: a float value to threshold to binarize instance border map.
            It turns uncertain area to 1 and other area to 0. Defaults to 0.4.
        radius: the radius of the disk-shaped footprint used in `opening`. Defaults to 2.
        min_object_size: objects smaller than this size (in pixel) are removed. Defaults to 10.
        postprocess_fn: additional post-process function on the markers.
            If not provided, :py:class:`monai.transforms.post.FillHoles()` will be used.
        dtype: target data type to convert to. Defaults to np.int64.

    """

    backend = [TransformBackends.NUMPY]

    def __init__(
        self,
        threshold: float = 0.4,
        radius: int = 2,
        min_object_size: int = 10,
        postprocess_fn: Callable | None = None,
        dtype: DtypeLike = np.int64,
    ) -> None:
        self.threshold = threshold
        self.radius = radius
        self.dtype = dtype
        if postprocess_fn is None:
            postprocess_fn = FillHoles()

        self.postprocess_fn = postprocess_fn
        self.remove_small_objects = RemoveSmallObjects(min_size=min_object_size) if min_object_size > 0 else None

    def __call__(self, mask: NdarrayOrTensor, instance_border: NdarrayOrTensor) -> NdarrayOrTensor:  # type: ignore
        """
        Args:
            mask: binary segmentation map, the output of :py:class:`GenerateWatershedMask`.
                Shape must be [1, H, W] or [H, W].
            instance_border: instance border map, the output of :py:class:`GenerateInstanceBorder`.
                Shape must be [1, H, W].
        """
        if len(mask.shape) == 3:
            if mask.shape[0] != 1:
                raise ValueError(f"The mask should have only one channel, but got {mask.shape[0]}.")
        elif len(mask.shape) == 2:
            mask = mask[None]
        else:
            raise ValueError(f"The mask should have the shape of [1, H, W] or [H, W], but got {mask.shape}.")
        if instance_border.shape[0] != 1 or instance_border.ndim != 3:
            raise ValueError(f"Input instance_border should be with size of [1, H, W], but got {instance_border.shape}")

        instance_border = instance_border >= self.threshold  # uncertain area

        marker = mask - convert_to_dst_type(instance_border, mask)[0]  # certain foreground
        marker[marker < 0] = 0  # type: ignore
        marker = self.postprocess_fn(marker)
        marker = convert_to_numpy(marker)

        marker = opening(marker.squeeze(), disk(self.radius))
        marker = label(marker)[0][None]
        if self.remove_small_objects is not None:
            marker = self.remove_small_objects(marker)

        return convert_to_dst_type(marker, mask, dtype=self.dtype)[0]


class GenerateSuccinctContour(Transform):
    """
    Converts SciPy-style contours (generated by skimage.measure.find_contours) to a more succinct version which only includes
    the pixels to which lines need to be drawn (i.e. not the intervening pixels along each line).

    Args:
        height: height of bounding box, used to detect direction of line segment.
        width: width of bounding box, used to detect direction of line segment.

    Returns:
        the pixels that need to be joined by straight lines to describe the outmost pixels of the foreground similar to
            OpenCV's cv.CHAIN_APPROX_SIMPLE (counterclockwise)
    """

    def __init__(self, height: int, width: int) -> None:
        self.height = height
        self.width = width

    def _generate_contour_coord(self, current: np.ndarray, previous: np.ndarray) -> tuple[int, int]:
        """
        Generate contour coordinates. Given the previous and current coordinates of border positions,
        returns the int pixel that marks the extremity of the segmented pixels.

        Args:
            current: coordinates of the current border position.
            previous: coordinates of the previous border position.
        """

        p_delta = (current[0] - previous[0], current[1] - previous[1])

        if p_delta in ((0.0, 1.0), (0.5, 0.5), (1.0, 0.0)):
            row = int(current[0] + 0.5)
            col = int(current[1])
        elif p_delta in ((0.0, -1.0), (0.5, -0.5)):
            row = int(current[0])
            col = int(current[1])
        elif p_delta in ((-1, 0.0), (-0.5, -0.5)):
            row = int(current[0])
            col = int(current[1] + 0.5)
        elif p_delta == (-0.5, 0.5):
            row = int(current[0] + 0.5)
            col = int(current[1] + 0.5)

        return row, col

    def _calculate_distance_from_top_left(self, sequence: Sequence[tuple[int, int]]) -> int:
        """
        Each sequence of coordinates describes a boundary between foreground and background starting and ending at two sides
        of the bounding box. To order the sequences correctly, we compute the distance from the top-left of the bounding box
        around the perimeter in a clockwise direction.

        Args:
            sequence: list of border points coordinates.

        Returns:
            the distance round the perimeter of the bounding box from the top-left origin
        """
        distance: int
        first_coord = sequence[0]
        if first_coord[0] == 0:
            distance = first_coord[1]
        elif first_coord[1] == self.width - 1:
            distance = self.width + first_coord[0]
        elif first_coord[0] == self.height - 1:
            distance = 2 * self.width + self.height - first_coord[1]
        else:
            distance = 2 * (self.width + self.height) - first_coord[0]

        return distance

    def __call__(self, contours: list[np.ndarray]) -> np.ndarray:
        """
        Args:
            contours: list of (n, 2)-ndarrays, scipy-style clockwise line segments, with lines separating foreground/background.
                Each contour is an ndarray of shape (n, 2), consisting of n (row, column) coordinates along the contour.
        """
        pixels: list[tuple[int, int]] = []
        sequences = []
        corners = [False, False, False, False]

        for group in contours:
            sequence: list[tuple[int, int]] = []
            last_added = None
            prev = None
            corner = -1

            for i, coord in enumerate(group):
                if i == 0:
                    # originating from the top, so must be heading south east
                    if coord[0] == 0.0:
                        corner = 1
                        pixel = (0, int(coord[1] - 0.5))
                        if pixel[1] == self.width - 1:
                            corners[1] = True
                        elif pixel[1] == 0.0:
                            corners[0] = True
                    # originating from the left, so must be heading north east
                    elif coord[1] == 0.0:
                        corner = 0
                        pixel = (int(coord[0] + 0.5), 0)
                    # originating from the bottom, so must be heading north west
                    elif coord[0] == self.height - 1:
                        corner = 3
                        pixel = (int(coord[0]), int(coord[1] + 0.5))
                        if pixel[1] == self.width - 1:
                            corners[2] = True
                    # originating from the right, so must be heading south west
                    elif coord[1] == self.width - 1:
                        corner = 2
                        pixel = (int(coord[0] - 0.5), int(coord[1]))
                    else:
                        warnings.warn(f"Invalid contour coord {coord} is generated, skip this instance.")
                        return None  # type: ignore
                    sequence.append(pixel)
                    last_added = pixel
                elif i == len(group) - 1:
                    # add this point
                    pixel = self._generate_contour_coord(coord, prev)  # type: ignore
                    if pixel != last_added:
                        sequence.append(pixel)
                        last_added = pixel
                elif np.any(coord - prev != group[i + 1] - coord):
                    pixel = self._generate_contour_coord(coord, prev)  # type: ignore
                    if pixel != last_added:
                        sequence.append(pixel)
                        last_added = pixel

                # flag whether each corner has been crossed
                if i == len(group) - 1:
                    if corner == 0:
                        if coord[0] == 0:
                            corners[corner] = True
                    elif corner == 1:
                        if coord[1] == self.width - 1:
                            corners[corner] = True
                    elif corner == 2:
                        if coord[0] == self.height - 1:
                            corners[corner] = True
                    elif corner == 3:
                        if coord[1] == 0.0:
                            corners[corner] = True

                prev = coord
            dist = self._calculate_distance_from_top_left(sequence)

            sequences.append({"distance": dist, "sequence": sequence})

        # check whether we need to insert any missing corners
        if corners[0] is False:
            sequences.append({"distance": 0, "sequence": [(0, 0)]})
        if corners[1] is False:
            sequences.append({"distance": self.width, "sequence": [(0, self.width - 1)]})
        if corners[2] is False:
            sequences.append({"distance": self.width + self.height, "sequence": [(self.height - 1, self.width - 1)]})
        if corners[3] is False:
            sequences.append({"distance": 2 * self.width + self.height, "sequence": [(self.height - 1, 0)]})

        # join the sequences into a single contour
        # starting at top left and rotating clockwise
        sequences.sort(key=lambda x: x.get("distance"))  # type: ignore

        last = (-1, -1)
        for _sequence in sequences:
            if _sequence["sequence"][0] == last:  # type: ignore
                pixels.pop()
            if pixels:
                pixels = [*pixels, *_sequence["sequence"]]  # type: ignore
            else:
                pixels = _sequence["sequence"]  # type: ignore
            last = pixels[-1]

        if pixels[0] == last:
            pixels.pop(0)

        if pixels[0] == (0, 0):
            pixels.append(pixels.pop(0))

        return np.flip(convert_to_numpy(pixels, dtype=np.int32))  # type: ignore


class GenerateInstanceContour(Transform):
    """
    Generate contour for each instance in a 2D array. Use `GenerateSuccinctContour` to only include
    the pixels to which lines need to be drawn

    Args:
        min_num_points: assumed that the created contour does not form a contour if it does not contain more points
            than the specified value. Defaults to 3.
        contour_level: an optional value for `skimage.measure.find_contours` to find contours in the array.
            If not provided, the level is set to `(max(image) + min(image)) / 2`.

    """

    backend = [TransformBackends.NUMPY]

    def __init__(self, min_num_points: int = 3, contour_level: float | None = None) -> None:
        self.contour_level = contour_level
        self.min_num_points = min_num_points

    def __call__(self, inst_mask: NdarrayOrTensor, offset: Sequence[int] | None = (0, 0)) -> np.ndarray | None:
        """
        Args:
            inst_mask: segmentation mask for a single instance. Shape should be [1, H, W, [D]]
            offset: optional offset of starting position of the instance mask in the original array. Default to 0 for each dim.
        """
        inst_mask = inst_mask.squeeze()  # squeeze channel dim
        inst_mask = convert_to_numpy(inst_mask)
        inst_contour_cv = find_contours(inst_mask, level=self.contour_level)
        generate_contour = GenerateSuccinctContour(inst_mask.shape[0], inst_mask.shape[1])
        inst_contour = generate_contour(inst_contour_cv)
        if inst_contour is None:
            return None
        # less than `self.min_num_points` points don't make a contour, so skip.
        # They are likely to be artifacts as the contours obtained via approximation.
        if inst_contour.shape[0] < self.min_num_points:
            print(f"< {self.min_num_points} points don't make a contour, so skipped!")
            return None
        # check for tricky shape
        elif len(inst_contour.shape) != 2:
            print(f"{len(inst_contour.shape)} != 2, check for tricky shapes!")
            return None
        else:
            inst_contour[:, 0] += offset[0]  # type: ignore
            inst_contour[:, 1] += offset[1]  # type: ignore
            return inst_contour


class GenerateInstanceCentroid(Transform):
    """
    Generate instance centroid using `skimage.measure.centroid`.

    Args:
        dtype: the data type of output centroid.

    """

    backend = [TransformBackends.NUMPY]

    def __init__(self, dtype: DtypeLike | None = int) -> None:
        self.dtype = dtype

    def __call__(self, inst_mask: NdarrayOrTensor, offset: Sequence[int] | int = 0) -> NdarrayOrTensor:
        """
        Args:
            inst_mask: segmentation mask for a single instance. Shape should be [1, H, W, [D]]
            offset: optional offset of starting position of the instance mask in the original array. Default to 0 for each dim.

        """
        inst_mask = convert_to_numpy(inst_mask)
        inst_mask = inst_mask.squeeze(0)  # squeeze channel dim
        ndim = len(inst_mask.shape)
        offset = ensure_tuple_rep(offset, ndim)

        inst_centroid = centroid(inst_mask)
        for i in range(ndim):
            inst_centroid[i] += offset[i]

        return convert_to_dst_type(inst_centroid, inst_mask, dtype=self.dtype)[0]


class GenerateInstanceType(Transform):
    """
    Generate instance type and probability for each instance.
    """

    backend = [TransformBackends.NUMPY]

    def __call__(  # type: ignore
        self, type_pred: NdarrayOrTensor, seg_pred: NdarrayOrTensor, bbox: np.ndarray, instance_id: int
    ) -> tuple[int, float]:
        """
        Args:
            type_pred: pixel-level type prediction map after activation function.
            seg_pred: pixel-level segmentation prediction map after activation function.
            bbox: bounding box coordinates of the instance, shape is [channel, 2 * spatial dims].
            instance_id: get instance type from specified instance id.
        """

        rmin, rmax, cmin, cmax = bbox.flatten()
        seg_map_crop = seg_pred[0, rmin:rmax, cmin:cmax]
        type_map_crop = type_pred[0, rmin:rmax, cmin:cmax]

        seg_map_crop = convert_to_dst_type(seg_map_crop == instance_id, type_map_crop, dtype=bool)[0]

        inst_type = type_map_crop[seg_map_crop]  # type: ignore
        type_list, type_pixels = unique(inst_type, return_counts=True)
        type_list = list(zip(type_list, type_pixels))
        type_list = sorted(type_list, key=lambda x: x[1], reverse=True)
        inst_type = type_list[0][0]
        if inst_type == 0:  # ! pick the 2nd most dominant if exist
            if len(type_list) > 1:
                inst_type = type_list[1][0]
        type_dict = {v[0]: v[1] for v in type_list}
        type_prob = type_dict[inst_type] / (sum(seg_map_crop) + 1.0e-6)

        return (int(inst_type), float(type_prob))


class HoVerNetInstanceMapPostProcessing(Transform):
    """
    The post-processing transform for HoVerNet model to generate instance segmentation map.
    It generates an instance segmentation map as well as a dictionary containing centroids, bounding boxes, and contours
    for each instance.

    Args:
        activation: the activation layer to be applied on the input probability map.
            It can be "softmax" or "sigmoid" string, or any callable. Defaults to "softmax".
        mask_threshold: a float value to threshold to binarize probability map to generate mask.
        min_object_size: objects smaller than this size (in pixel) are removed. Defaults to 10.
        sobel_kernel_size: the size of the Sobel kernel used in :py:class:`GenerateInstanceBorder`. Defaults to 5.
        distance_smooth_fn: smoothing function for distance map.
            If not provided, :py:class:`monai.transforms.intensity.GaussianSmooth()` will be used.
        marker_threshold: a float value to threshold to binarize instance border map for markers.
            It turns uncertain area to 1 and other area to 0. Defaults to 0.4.
        marker_radius: the radius of the disk-shaped footprint used in `opening` of markers. Defaults to 2.
        marker_postprocess_fn: post-process function for watershed markers.
            If not provided, :py:class:`monai.transforms.post.FillHoles()` will be used.
        watershed_connectivity: `connectivity` argument of `skimage.segmentation.watershed`.
        min_num_points: minimum number of points to be considered as a contour. Defaults to 3.
        contour_level: an optional value for `skimage.measure.find_contours` to find contours in the array.
            If not provided, the level is set to `(max(image) + min(image)) / 2`.
        device: target device to put the output Tensor data.
    """

    def __init__(
        self,
        activation: str | Callable = "softmax",
        mask_threshold: float | None = None,
        min_object_size: int = 10,
        sobel_kernel_size: int = 5,
        distance_smooth_fn: Callable | None = None,
        marker_threshold: float = 0.4,
        marker_radius: int = 2,
        marker_postprocess_fn: Callable | None = None,
        watershed_connectivity: int | None = 1,
        min_num_points: int = 3,
        contour_level: float | None = None,
        device: str | torch.device | None = None,
    ) -> None:
        super().__init__()
        self.device = device
        self.generate_watershed_mask = GenerateWatershedMask(
            activation=activation, threshold=mask_threshold, min_object_size=min_object_size
        )
        self.generate_instance_border = GenerateInstanceBorder(kernel_size=sobel_kernel_size)
        self.generate_distance_map = GenerateDistanceMap(smooth_fn=distance_smooth_fn)
        self.generate_watershed_markers = GenerateWatershedMarkers(
            threshold=marker_threshold,
            radius=marker_radius,
            postprocess_fn=marker_postprocess_fn,
            min_object_size=min_object_size,
        )
        self.watershed = Watershed(connectivity=watershed_connectivity)
        self.generate_instance_contour = GenerateInstanceContour(
            min_num_points=min_num_points, contour_level=contour_level
        )
        self.generate_instance_centroid = GenerateInstanceCentroid()

    def __call__(  # type: ignore
        self, nuclear_prediction: NdarrayOrTensor, hover_map: NdarrayOrTensor
    ) -> tuple[dict, NdarrayOrTensor]:
        """post-process instance segmentation branches (NP and HV) to generate instance segmentation map.

        Args:
            nuclear_prediction: the output of NP (nuclear prediction) branch of HoVerNet model
            hover_map: the output of HV (hover map) branch of HoVerNet model
        """

        # Process NP and HV branch using watershed algorithm
        watershed_mask = self.generate_watershed_mask(nuclear_prediction)
        instance_borders = self.generate_instance_border(watershed_mask, hover_map)
        distance_map = self.generate_distance_map(watershed_mask, instance_borders)
        watershed_markers = self.generate_watershed_markers(watershed_mask, instance_borders)
        instance_map = self.watershed(distance_map, watershed_mask, watershed_markers)

        # Create bounding boxes, contours and centroids
        instance_ids = set(np.unique(instance_map)) - {0}  # exclude background
        instance_info = {}
        for inst_id in instance_ids:
            instance_mask = instance_map == inst_id
            instance_bbox = BoundingRect()(instance_mask)

            instance_mask = instance_mask[
                :, instance_bbox[0][0] : instance_bbox[0][1], instance_bbox[0][2] : instance_bbox[0][3]
            ]
            offset = [instance_bbox[0][2], instance_bbox[0][0]]
            instance_contour = self.generate_instance_contour(FillHoles()(instance_mask), offset)
            if instance_contour is not None:
                instance_centroid = self.generate_instance_centroid(instance_mask, offset)
                instance_info[inst_id] = {
                    "bounding_box": instance_bbox,
                    "centroid": instance_centroid,
                    "contour": instance_contour,
                }
        instance_map = convert_to_tensor(instance_map, device=self.device)
        return instance_info, instance_map


class HoVerNetNuclearTypePostProcessing(Transform):
    """
    The post-processing transform for HoVerNet model to generate nuclear type information.
    It updates the input instance info dictionary with information about types of the nuclei (value and probability).
    Also if requested (`return_type_map=True`), it generates a pixel-level type map.

    Args:
        activation: the activation layer to be applied on nuclear type branch. It can be "softmax" or "sigmoid" string,
            or any callable. Defaults to "softmax".
        threshold: an optional float value to threshold to binarize probability map.
            If not provided, defaults to 0.5 when activation is not "softmax", otherwise None.
        return_type_map: whether to calculate and return pixel-level type map.
        device: target device to put the output Tensor data.

    """

    def __init__(
        self,
        activation: str | Callable = "softmax",
        threshold: float | None = None,
        return_type_map: bool = True,
        device: str | torch.device | None = None,
    ) -> None:
        super().__init__()
        self.device = device
        self.return_type_map = return_type_map
        self.generate_instance_type = GenerateInstanceType()

        # set activation layer
        use_softmax = False
        use_sigmoid = False
        activation_fn = None
        if isinstance(activation, str):
            if activation.lower() == "softmax":
                use_softmax = True
            elif activation.lower() == "sigmoid":
                use_sigmoid = True
            else:
                raise ValueError(
                    f"The activation should be 'softmax' or 'sigmoid' string, or any callable. '{activation}' was given."
                )
        elif callable(activation):
            activation_fn = activation
        else:
            raise ValueError(f"The activation type should be either str or callable. '{type(activation)}' was given.")
        self.activation = Activations(softmax=use_softmax, sigmoid=use_sigmoid, other=activation_fn)

        # set discretization transform
        if not use_softmax and threshold is None:
            threshold = 0.5
        self.as_discrete = AsDiscrete(threshold=threshold, argmax=use_softmax)

    def __call__(  # type: ignore
        self, type_prediction: NdarrayOrTensor, instance_info: dict[int, dict], instance_map: NdarrayOrTensor
    ) -> tuple[dict, NdarrayOrTensor | None]:
        """Process NC (type prediction) branch and combine it with instance segmentation
        It updates the instance_info with instance type and associated probability, and generate instance type map.

        Args:
            instance_info: instance information dictionary, the output of :py:class:`HoVerNetInstanceMapPostProcessing`
            instance_map: instance segmentation map, the output of :py:class:`HoVerNetInstanceMapPostProcessing`
            type_prediction: the output of NC (type prediction) branch of HoVerNet model
        """
        type_prediction = self.activation(type_prediction)
        type_prediction = self.as_discrete(type_prediction)

        type_map = None
        if self.return_type_map:
            type_map = convert_to_dst_type(torch.zeros(instance_map.shape), instance_map)[0]

        for inst_id in instance_info:
            instance_type, instance_type_prob = self.generate_instance_type(
                type_pred=type_prediction,
                seg_pred=instance_map,
                bbox=instance_info[inst_id]["bounding_box"],
                instance_id=inst_id,
            )
            # update instance info dict with type data
            instance_info[inst_id]["type_prob"] = instance_type_prob
            instance_info[inst_id]["type"] = instance_type

            # update instance type map
            if type_map is not None:
                type_map[instance_map == inst_id] = instance_type
                type_map = convert_to_tensor(type_map, device=self.device)

        return instance_info, type_map



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.optimizers.lr_scheduler import WarmupCosineSchedule


class SchedulerTestNet(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(1, 1, 1)
        self.conv2 = torch.nn.Conv2d(1, 1, 1)

    def forward(self, x):
        return self.conv2(torch.nn.functional.relu(self.conv1(x)))


TEST_CASE_LRSCHEDULER = [
    [{"warmup_steps": 2, "t_total": 10}, [0.000, 0.500, 1.00, 0.962, 0.854, 0.691, 0.500, 0.309, 0.146, 0.038]],
    [
        {"warmup_steps": 2, "t_total": 10, "warmup_multiplier": 0.1},
        [0.1, 0.55, 1.00, 0.962, 0.854, 0.691, 0.500, 0.309, 0.146, 0.038],
    ],
    [
        {"warmup_steps": 2, "t_total": 10, "warmup_multiplier": 0.1, "end_lr": 0.309},
        [0.1, 0.55, 1.00, 0.962, 0.854, 0.691, 0.500, 0.309, 0.309, 0.309],
    ],
]


class TestLRSCHEDULER(unittest.TestCase):
    @parameterized.expand(TEST_CASE_LRSCHEDULER)
    def test_shape(self, input_param, expected_lr):
        net = SchedulerTestNet()
        optimizer = torch.optim.Adam(net.parameters(), lr=1.0)
        scheduler = WarmupCosineSchedule(optimizer, **input_param)
        self.assertEqual(len([scheduler.get_last_lr()[0]]), 1)
        lrs_1 = []
        for _ in range(input_param["t_total"]):
            lrs_1.append(float(f"{scheduler.get_last_lr()[0]:.3f}"))
            optimizer.step()
            scheduler.step()
        for a, b in zip(lrs_1, expected_lr):
            self.assertEqual(a, b, msg=f"LR is wrong ! expected {b}, got {a}")

    def test_error(self):
        """Should fail because warmup_multiplier is outside 0..1"""
        net = SchedulerTestNet()
        optimizer = torch.optim.Adam(net.parameters(), lr=1.0)
        with self.assertRaises(ValueError):
            WarmupCosineSchedule(optimizer, warmup_steps=2, t_total=10, warmup_multiplier=-1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import unittest

import nibabel as nib
import numpy as np
import torch
from ignite.engine import Engine, Events
from torch.utils.data import DataLoader

from monai.data import ImageDataset, create_test_image_3d
from monai.inferers import sliding_window_inference
from monai.networks import eval_mode, predict_segmentation
from monai.networks.nets import UNet
from monai.transforms import EnsureChannelFirst, SaveImage
from monai.utils import pytorch_after, set_determinism
from tests.utils import DistTestCase, TimedCall, make_nifti_image, skip_if_quick


def run_test(batch_size, img_name, seg_name, output_dir, device="cuda:0"):
    ds = ImageDataset(
        [img_name],
        [seg_name],
        transform=EnsureChannelFirst(channel_dim="no_channel"),
        seg_transform=EnsureChannelFirst(channel_dim="no_channel"),
        image_only=True,
    )
    loader = DataLoader(ds, batch_size=1, pin_memory=torch.cuda.is_available())

    net = UNet(
        spatial_dims=3, in_channels=1, out_channels=1, channels=(4, 8, 16, 32), strides=(2, 2, 2), num_res_units=2
    ).to(device)
    roi_size = (16, 32, 48)
    sw_batch_size = batch_size

    saver = SaveImage(output_dir=output_dir, output_ext=".nii.gz", output_postfix="seg")

    def _sliding_window_processor(_engine, batch):
        img = batch[0]  # first item from ImageDataset is the input image
        with eval_mode(net):
            seg_probs = sliding_window_inference(img.to(device), roi_size, sw_batch_size, net, device=device)
            return predict_segmentation(seg_probs)

    def save_func(engine):
        if pytorch_after(1, 9, 1):
            for m in engine.state.output:
                saver(m)
        else:
            saver(engine.state.output[0])

    infer_engine = Engine(_sliding_window_processor)
    infer_engine.add_event_handler(Events.ITERATION_COMPLETED, save_func)
    infer_engine.run(loader)

    basename = os.path.basename(img_name)[: -len(".nii.gz")]
    saved_name = os.path.join(output_dir, basename, f"{basename}_seg.nii.gz")
    return saved_name


@skip_if_quick
class TestIntegrationSlidingWindow(DistTestCase):
    def setUp(self):
        set_determinism(seed=0)

        im, seg = create_test_image_3d(28, 25, 63, rad_max=10, noise_max=1, num_objs=4, num_seg_classes=1)
        self.img_name = make_nifti_image(im)
        self.seg_name = make_nifti_image(seg)
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu:0")

    def tearDown(self):
        set_determinism(seed=None)
        if os.path.exists(self.img_name):
            os.remove(self.img_name)
        if os.path.exists(self.seg_name):
            os.remove(self.seg_name)

    @TimedCall(seconds=20)
    def test_training(self):
        set_determinism(seed=0)
        with tempfile.TemporaryDirectory() as tempdir:
            output_file = run_test(
                batch_size=2, img_name=self.img_name, seg_name=self.seg_name, output_dir=tempdir, device=self.device
            )
            output_image = nib.load(output_file).get_fdata()
            np.testing.assert_allclose(np.sum(output_image), 33621)
            np.testing.assert_allclose(output_image.shape, (28, 25, 63))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import unittest

from monai.apps import CrossValidation, DecathlonDataset
from monai.data import MetaTensor
from monai.transforms import Compose, EnsureChannelFirstd, LoadImaged, ScaleIntensityd
from tests.utils import skip_if_downloading_fails, skip_if_quick


class TestCrossValidation(unittest.TestCase):
    @skip_if_quick
    def test_values(self):
        testing_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "testing_data")
        train_transform = Compose(
            [
                LoadImaged(keys=["image", "label"]),
                EnsureChannelFirstd(keys=["image", "label"], channel_dim="no_channel"),
                ScaleIntensityd(keys="image"),
            ]
        )
        val_transform = LoadImaged(keys=["image", "label"])

        def _test_dataset(dataset):
            self.assertEqual(len(dataset), 52)
            self.assertTrue("image" in dataset[0])
            self.assertTrue("label" in dataset[0])
            self.assertTrue(isinstance(dataset[0]["image"], MetaTensor))
            self.assertTupleEqual(dataset[0]["image"].shape, (1, 34, 49, 41))

        cvdataset = CrossValidation(
            dataset_cls=DecathlonDataset,
            nfolds=5,
            seed=12345,
            root_dir=testing_dir,
            task="Task04_Hippocampus",
            section="validation",
            transform=train_transform,
            download=True,
        )

        with skip_if_downloading_fails():
            data = cvdataset.get_dataset(folds=0)

        _test_dataset(data)

        # test training data for fold [1, 2, 3, 4] of 5 splits
        data = cvdataset.get_dataset(folds=[1, 2, 3, 4])
        self.assertTupleEqual(data[0]["image"].shape, (1, 35, 52, 33))
        self.assertEqual(len(data), 208)
        # test train / validation for fold 4 of 5 splits
        data = cvdataset.get_dataset(folds=[4], transform=val_transform, download=False)
        # val_transform doesn't add the channel dim to shape
        self.assertTupleEqual(data[0]["image"].shape, (38, 53, 30))
        self.assertEqual(len(data), 52)
        data = cvdataset.get_dataset(folds=[0, 1, 2, 3])
        self.assertTupleEqual(data[0]["image"].shape, (1, 34, 49, 41))
        self.assertEqual(len(data), 208)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools
import os
import random
import shutil
import string
import tempfile
import unittest

import nibabel as nib
import numpy as np
import torch
from parameterized import parameterized

from monai.data import MetaTensor
from monai.data.image_reader import ITKReader, NibabelReader
from monai.data.image_writer import ITKWriter
from monai.transforms import Compose, EnsureChannelFirstd, LoadImaged, ResampleToMatch, SaveImage, SaveImaged
from monai.utils import optional_import
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.utils import assert_allclose, download_url_or_skip_test, testing_data_config

_, has_itk = optional_import("itk", allow_namespace_pkg=True)

TEST_CASES = ["itkreader", "nibabelreader"]


def get_rand_fname(len=10, suffix=".nii.gz"):
    letters = string.ascii_letters
    out = "".join(random.choice(letters) for _ in range(len))
    out += suffix
    return out


@unittest.skipUnless(has_itk, "itk not installed")
class TestResampleToMatch(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(__class__, cls).setUpClass()
        cls.fnames = []
        cls.tmpdir = tempfile.mkdtemp()
        for key in ("0000_t2_tse_tra_4", "0000_ep2d_diff_tra_7"):
            fname = os.path.join(cls.tmpdir, f"test_{key}.nii.gz")
            url = testing_data_config("images", key, "url")
            hash_type = testing_data_config("images", key, "hash_type")
            hash_val = testing_data_config("images", key, "hash_val")
            download_url_or_skip_test(url=url, filepath=fname, hash_type=hash_type, hash_val=hash_val)
            cls.fnames.append(fname)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)
        super(__class__, cls).tearDownClass()

    @parameterized.expand(itertools.product([NibabelReader, ITKReader], ["monai.data.NibabelWriter", ITKWriter]))
    def test_correct(self, reader, writer):
        loader = Compose([LoadImaged(("im1", "im2"), reader=reader), EnsureChannelFirstd(("im1", "im2"))])
        data = loader({"im1": self.fnames[0], "im2": self.fnames[1]})
        tr = ResampleToMatch()
        im_mod = tr(data["im2"], data["im1"])

        # check lazy resample
        call_param = {"img": data["im2"], "img_dst": data["im1"]}
        test_resampler_lazy(tr, im_mod, init_param={}, call_param=call_param)

        saver = SaveImaged(
            "im3", output_dir=self.tmpdir, output_postfix="", separate_folder=False, writer=writer, resample=False
        )
        im_mod.meta["filename_or_obj"] = get_rand_fname()
        saver({"im3": im_mod})

        saved = nib.load(os.path.join(self.tmpdir, im_mod.meta["filename_or_obj"]))
        assert_allclose(data["im1"].shape[1:], saved.shape)
        assert_allclose(saved.header["dim"][:4], np.array([3, 384, 384, 19]))

    def test_inverse(self):
        loader = Compose([LoadImaged(("im1", "im2"), image_only=True), EnsureChannelFirstd(("im1", "im2"))])
        data = loader({"im1": self.fnames[0], "im2": self.fnames[1]})
        tr = ResampleToMatch()
        im_mod = tr(data["im2"], data["im1"])
        self.assertNotEqual(im_mod.shape, data["im2"].shape)
        self.assertGreater(((im_mod.affine - data["im2"].affine) ** 2).sum() ** 0.5, 1e-2)
        # inverse
        im_mod2 = tr.inverse(im_mod)
        self.assertEqual(im_mod2.shape, data["im2"].shape)
        self.assertLess(((im_mod2.affine - data["im2"].affine) ** 2).sum() ** 0.5, 1e-2)
        self.assertEqual(im_mod2.applied_operations, [])

    def test_no_name(self):
        img_1 = MetaTensor(torch.zeros(1, 2, 2, 2))
        img_2 = MetaTensor(torch.zeros(1, 3, 3, 3))
        im_mod = ResampleToMatch()(img_1, img_2)
        self.assertEqual(im_mod.meta["filename_or_obj"], "resample_to_match_source")
        SaveImage(output_dir=self.tmpdir, output_postfix="", separate_folder=False, resample=False)(im_mod)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import unittest

import torch

import monai.transforms as mt
from monai.data.dataloader import DataLoader
from monai.data.video_dataset import CameraDataset, VideoDataset, VideoFileDataset
from monai.utils.module import optional_import
from tests.utils import assert_allclose, download_url_or_skip_test, testing_data_config

cv2, has_cv2 = optional_import("cv2")

NUM_CAPTURE_DEVICES = CameraDataset.get_num_devices()
TRANSFORMS = mt.Compose(
    [mt.EnsureChannelFirst(True, "no_channel"), mt.DivisiblePad(16), mt.ScaleIntensity(), mt.CastToType(torch.float32)]
)


class Base:
    class TestVideoDataset(unittest.TestCase):
        video_source: int | str
        ds: type[VideoDataset]

        def get_video_source(self):
            return self.video_source

        def get_ds(self, *args, **kwargs) -> VideoDataset:
            return self.ds(video_source=self.get_video_source(), transform=TRANSFORMS, *args, **kwargs)  # type: ignore

        @unittest.skipIf(has_cv2, "Only tested when OpenCV not installed.")
        def test_no_opencv_raises(self):
            with self.assertRaises(RuntimeError):
                _ = self.get_ds(max_num_frames=10)

        @unittest.skipUnless(has_cv2, "OpenCV required.")
        def test_multiprocessing(self):
            for num_workers in (0, 2):
                multiprocessing = num_workers > 0
                ds = self.get_ds(max_num_frames=100, multiprocessing=multiprocessing)
                dl = DataLoader(ds, num_workers=num_workers, batch_size=2)
                _ = next(iter(dl))

        @unittest.skipUnless(has_cv2, "OpenCV required.")
        def test_multiple_sources(self, should_match: bool = True):
            ds1 = self.get_ds()
            ds2 = self.get_ds()
            if should_match:
                assert_allclose(ds1.get_frame(), ds2.get_frame())

        @unittest.skipUnless(has_cv2, "OpenCV required.")
        def test_dataset(self, known_num_frames=None, known_fps=None):
            num_frames = (10,) if known_num_frames is None else (10, None)
            for max_num_frames in num_frames:
                ds = self.get_ds(max_num_frames=max_num_frames)
                if known_fps is not None:
                    self.assertEqual(ds.get_fps(), known_fps)
                frames = list(ds)
                if max_num_frames is not None:
                    self.assertEqual(len(frames), max_num_frames)
                elif known_num_frames is not None:
                    self.assertEqual(len(frames), len(ds))
                for f in frames:
                    self.assertTupleEqual(f.shape, frames[0].shape)


@unittest.skipIf(NUM_CAPTURE_DEVICES == 0, "At least one capture device required.")
class TestCameraDataset(Base.TestVideoDataset):
    video_source = 0
    ds = CameraDataset

    @unittest.skipUnless(has_cv2, "OpenCV required.")
    def test_multiple_sources(self):
        super().test_multiple_sources(should_match=False)

    @unittest.skipUnless(has_cv2, "OpenCV required.")
    def test_device_out_of_range(self):
        capture_device = NUM_CAPTURE_DEVICES + 1
        with self.assertRaises(RuntimeError):
            _ = CameraDataset(capture_device, TRANSFORMS, 0)


class TestVideoFileDataset(Base.TestVideoDataset):
    ds = VideoFileDataset

    @classmethod
    def setUpClass(cls):
        super(__class__, cls).setUpClass()
        codecs = VideoFileDataset.get_available_codecs()
        if ".mp4" in codecs.values():
            fname = "endo.mp4"
            config = testing_data_config("videos", "endovis")
            cls.known_fps = 2.0
            cls.known_num_frames = 23
        elif ".avi" in codecs.values():
            fname = "ultrasound.avi"
            config = testing_data_config("videos", "ultrasound")
            cls.known_fps = 2.0
            cls.known_num_frames = 523
        else:
            cls.known_fps = None
            cls.known_num_frames = None
            cls.video_source = None
            return
        cls.video_source = os.path.join(os.path.dirname(__file__), "testing_data", fname)
        download_url_or_skip_test(
            url=config["url"],
            filepath=cls.video_source,
            hash_val=config.get("hash_val"),
            hash_type=config.get("hash_type", "sha256"),
        )

    @unittest.skipUnless(has_cv2, "OpenCV required.")
    def test_dataset(self):
        super().test_dataset(self.known_num_frames, self.known_fps)
        self.assertEqual(self.get_ds().get_num_frames(), self.known_num_frames)

    def test_available_codecs(self):
        codecs = VideoFileDataset.get_available_codecs()
        if not has_cv2:
            self.assertEqual(codecs, {})
        else:
            self.assertGreaterEqual(len(codecs), 0)

    def get_video_source(self):
        if self.video_source is None:
            raise unittest.SkipTest("missing required codecs")
        return super().get_video_source()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.fl.utils.constants import WeightType
from monai.fl.utils.exchange_object import ExchangeObject
from monai.utils.module import optional_import
from tests.utils import SkipIfNoModule

models, has_torchvision = optional_import("torchvision.models")

TEST_INIT_1 = [{"weights": None, "optim": None, "metrics": None, "weight_type": None, "statistics": None}, "{}"]
TEST_INIT_2: list = []
if has_torchvision:
    network = models.resnet18()
    TEST_INIT_2.append(
        {
            "weights": network.state_dict(),
            "optim": torch.optim.Adam(lr=1, params=network.parameters()).state_dict(),
            "metrics": {"accuracy": 1},
            "weight_type": WeightType.WEIGHT_DIFF,
            "statistics": {"some_stat": 1},
        }
    )
    TEST_INIT_2.append("{'weights': 122, 'optim': 2, 'metrics': 1, 'weight_type': fl_weight_diff, 'statistics': 1}")

TEST_FAILURE_METRICS = [{"weights": None, "optim": None, "metrics": 1, "weight_type": None, "statistics": None}]
TEST_FAILURE_STATISTICS = [{"weights": None, "optim": None, "metrics": None, "weight_type": None, "statistics": 1}]
TEST_FAILURE_WEIGHT_TYPE = [{"weights": None, "optim": None, "metrics": None, "weight_type": 1, "statistics": None}]


@SkipIfNoModule("torchvision")
@SkipIfNoModule("ignite")
class TestFLExchangeObject(unittest.TestCase):
    @parameterized.expand([TEST_INIT_1, TEST_INIT_2])
    def test_init(self, input_params, expected_str):
        eo = ExchangeObject(**input_params)
        self.assertIsInstance(eo, ExchangeObject)
        eo.summary()
        self.assertEqual(repr(eo), expected_str)

    @parameterized.expand([TEST_FAILURE_METRICS, TEST_FAILURE_STATISTICS, TEST_FAILURE_WEIGHT_TYPE])
    def test_failures(self, input_params):
        with self.assertRaises(ValueError):
            ExchangeObject(**input_params)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from .post.array import (
    GenerateDistanceMap,
    GenerateInstanceBorder,
    GenerateInstanceCentroid,
    GenerateInstanceContour,
    GenerateInstanceType,
    GenerateSuccinctContour,
    GenerateWatershedMarkers,
    GenerateWatershedMask,
    HoVerNetInstanceMapPostProcessing,
    HoVerNetNuclearTypePostProcessing,
    Watershed,
)
from .post.dictionary import (
    GenerateDistanceMapD,
    GenerateDistanceMapd,
    GenerateDistanceMapDict,
    GenerateInstanceBorderD,
    GenerateInstanceBorderd,
    GenerateInstanceBorderDict,
    GenerateInstanceCentroidD,
    GenerateInstanceCentroidd,
    GenerateInstanceCentroidDict,
    GenerateInstanceContourD,
    GenerateInstanceContourd,
    GenerateInstanceContourDict,
    GenerateInstanceTypeD,
    GenerateInstanceTyped,
    GenerateInstanceTypeDict,
    GenerateSuccinctContourD,
    GenerateSuccinctContourd,
    GenerateSuccinctContourDict,
    GenerateWatershedMarkersD,
    GenerateWatershedMarkersd,
    GenerateWatershedMarkersDict,
    GenerateWatershedMaskD,
    GenerateWatershedMaskd,
    GenerateWatershedMaskDict,
    HoVerNetInstanceMapPostProcessingD,
    HoVerNetInstanceMapPostProcessingd,
    HoVerNetInstanceMapPostProcessingDict,
    HoVerNetNuclearTypePostProcessingD,
    HoVerNetNuclearTypePostProcessingd,
    HoVerNetNuclearTypePostProcessingDict,
    WatershedD,
    Watershedd,
    WatershedDict,
)
from .stain.array import ExtractHEStains, NormalizeHEStains
from .stain.dictionary import (
    ExtractHEStainsd,
    ExtractHEStainsD,
    ExtractHEStainsDict,
    NormalizeHEStainsd,
    NormalizeHEStainsD,
    NormalizeHEStainsDict,
)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch

from monai.transforms import UltrasoundConfidenceMapTransform
from tests.utils import assert_allclose

TEST_INPUT = np.array(
    [
        [1, 2, 3, 23, 13, 22, 5, 1, 2, 3],
        [1, 2, 3, 12, 4, 6, 9, 1, 2, 3],
        [1, 2, 3, 8, 7, 10, 11, 1, 2, 3],
        [1, 2, 3, 14, 15, 16, 17, 1, 2, 3],
        [1, 2, 3, 18, 19, 20, 21, 1, 2, 3],
        [1, 2, 3, 24, 25, 26, 27, 1, 2, 3],
        [1, 2, 3, 28, 29, 30, 31, 1, 2, 3],
        [1, 2, 3, 32, 33, 34, 35, 1, 2, 3],
        [1, 2, 3, 36, 37, 38, 39, 1, 2, 3],
        [1, 2, 3, 40, 41, 42, 43, 1, 2, 3],
    ]
)

TEST_MASK = np.array(
    [
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [1, 1, 1, 0, 0, 0, 1, 1, 1, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
    ]
)

SINK_ALL_OUTPUT = np.array(
    [
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        [
            0.97514489,
            0.96762971,
            0.96164186,
            0.95463443,
            0.9941512,
            0.99023054,
            0.98559401,
            0.98230057,
            0.96601224,
            0.95119599,
        ],
        [
            0.92960533,
            0.92638451,
            0.9056675,
            0.9487176,
            0.9546961,
            0.96165853,
            0.96172303,
            0.92686401,
            0.92122613,
            0.89957239,
        ],
        [
            0.86490963,
            0.85723665,
            0.83798141,
            0.90816201,
            0.90816097,
            0.90815301,
            0.9081427,
            0.85933627,
            0.85146935,
            0.82948586,
        ],
        [
            0.77430346,
            0.76731372,
            0.74372311,
            0.89128774,
            0.89126885,
            0.89125066,
            0.89123521,
            0.76858589,
            0.76106647,
            0.73807776,
        ],
        [
            0.66098109,
            0.65327697,
            0.63090644,
            0.33086588,
            0.3308383,
            0.33081937,
            0.33080718,
            0.6557468,
            0.64825099,
            0.62593375,
        ],
        [
            0.52526945,
            0.51832586,
            0.49709412,
            0.25985059,
            0.25981009,
            0.25977729,
            0.25975222,
            0.52118958,
            0.51426328,
            0.49323164,
        ],
        [
            0.3697845,
            0.36318971,
            0.34424661,
            0.17386804,
            0.17382046,
            0.17377993,
            0.17374668,
            0.36689317,
            0.36036096,
            0.3415582,
        ],
        [
            0.19546374,
            0.1909659,
            0.17319999,
            0.08423318,
            0.08417993,
            0.08413242,
            0.08409104,
            0.19393909,
            0.18947485,
            0.17185031,
        ],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ]
)

SINK_MID_OUTPUT = np.array(
    [
        [
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
        ],
        [
            9.99996103e-01,
            9.99994823e-01,
            9.99993550e-01,
            9.99930863e-01,
            9.99990782e-01,
            9.99984683e-01,
            9.99979000e-01,
            9.99997804e-01,
            9.99995985e-01,
            9.99994325e-01,
        ],
        [
            9.99989344e-01,
            9.99988600e-01,
            9.99984099e-01,
            9.99930123e-01,
            9.99926598e-01,
            9.99824297e-01,
            9.99815032e-01,
            9.99991228e-01,
            9.99990881e-01,
            9.99988462e-01,
        ],
        [
            9.99980787e-01,
            9.99979264e-01,
            9.99975828e-01,
            9.59669286e-01,
            9.59664779e-01,
            9.59656566e-01,
            9.59648332e-01,
            9.99983882e-01,
            9.99983038e-01,
            9.99980732e-01,
        ],
        [
            9.99970181e-01,
            9.99969032e-01,
            9.99965730e-01,
            9.45197806e-01,
            9.45179593e-01,
            9.45163629e-01,
            9.45151458e-01,
            9.99973352e-01,
            9.99973254e-01,
            9.99971098e-01,
        ],
        [
            9.99958608e-01,
            9.99957307e-01,
            9.99953444e-01,
            4.24743523e-01,
            4.24713305e-01,
            4.24694646e-01,
            4.24685271e-01,
            9.99960948e-01,
            9.99961829e-01,
            9.99960347e-01,
        ],
        [
            9.99946675e-01,
            9.99945139e-01,
            9.99940312e-01,
            3.51353224e-01,
            3.51304003e-01,
            3.51268260e-01,
            3.51245366e-01,
            9.99947688e-01,
            9.99950165e-01,
            9.99949512e-01,
        ],
        [
            9.99935877e-01,
            9.99934088e-01,
            9.99928982e-01,
            2.51197134e-01,
            2.51130273e-01,
            2.51080014e-01,
            2.51045852e-01,
            9.99936187e-01,
            9.99939716e-01,
            9.99940022e-01,
        ],
        [
            9.99927846e-01,
            9.99925911e-01,
            9.99920188e-01,
            1.31550973e-01,
            1.31462736e-01,
            1.31394558e-01,
            1.31346069e-01,
            9.99927275e-01,
            9.99932142e-01,
            9.99933313e-01,
        ],
        [
            9.99924204e-01,
            9.99922004e-01,
            9.99915767e-01,
            3.04861147e-04,
            1.95998056e-04,
            0.00000000e00,
            2.05182682e-05,
            9.99923115e-01,
            9.99928835e-01,
            9.99930535e-01,
        ],
    ]
)

SINK_MIN_OUTPUT = np.array(
    [
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        [
            0.99997545,
            0.99996582,
            0.99995245,
            0.99856594,
            0.99898314,
            0.99777223,
            0.99394423,
            0.98588113,
            0.97283215,
            0.96096504,
        ],
        [
            0.99993872,
            0.99993034,
            0.9998832,
            0.9986147,
            0.99848741,
            0.9972981,
            0.99723719,
            0.94157173,
            0.9369832,
            0.91964243,
        ],
        [
            0.99990802,
            0.99989475,
            0.99986873,
            0.98610197,
            0.98610047,
            0.98609749,
            0.98609423,
            0.88741275,
            0.88112911,
            0.86349156,
        ],
        [
            0.99988924,
            0.99988509,
            0.99988698,
            0.98234089,
            0.98233591,
            0.98233065,
            0.98232562,
            0.81475172,
            0.80865978,
            0.79033138,
        ],
        [
            0.99988418,
            0.99988484,
            0.99988323,
            0.86796555,
            0.86795874,
            0.86795283,
            0.86794756,
            0.72418193,
            0.71847704,
            0.70022037,
        ],
        [
            0.99988241,
            0.99988184,
            0.99988103,
            0.85528225,
            0.85527303,
            0.85526389,
            0.85525499,
            0.61716519,
            0.61026209,
            0.59503671,
        ],
        [
            0.99988015,
            0.99987985,
            0.99987875,
            0.84258114,
            0.84257121,
            0.84256042,
            0.84254897,
            0.48997924,
            0.49083978,
            0.46891561,
        ],
        [
            0.99987865,
            0.99987827,
            0.9998772,
            0.83279589,
            0.83278624,
            0.83277384,
            0.83275897,
            0.36345545,
            0.33690244,
            0.35696828,
        ],
        [
            0.99987796,
            0.99987756,
            0.99987643,
            0.82873223,
            0.82872648,
            0.82871803,
            0.82870711,
            0.0,
            0.26106012,
            0.29978657,
        ],
    ]
)

SINK_MASK_OUTPUT = np.array(
    [
        [
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
            1.00000000e00,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            2.86416400e-01,
            7.93271181e-01,
            5.81341234e-01,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            1.98395623e-01,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            2.66733297e-01,
            2.80741490e-01,
            4.14078784e-02,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            7.91676486e-04,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            1.86244537e-04,
            1.53413401e-04,
            7.85806495e-05,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            5.09797387e-06,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            9.62904581e-07,
            7.23946225e-07,
            3.68824440e-07,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            4.79525316e-08,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            1.50939343e-10,
            1.17724874e-10,
            6.21760843e-11,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            6.08922784e-10,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            2.57593754e-13,
            1.94066716e-13,
            9.83784370e-14,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            9.80828665e-12,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            4.22323494e-16,
            3.17556633e-16,
            1.60789400e-16,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            1.90789819e-13,
        ],
        [
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            7.72677888e-19,
            5.83029424e-19,
            2.95946659e-19,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            4.97038275e-15,
        ],
        [
            2.71345908e-24,
            5.92006757e-24,
            2.25580089e-23,
            3.82601970e-18,
            3.82835349e-18,
            3.83302158e-18,
            3.84002606e-18,
            8.40760586e-16,
            1.83433696e-15,
            1.11629633e-15,
        ],
    ]
)


class TestUltrasoundConfidenceMapTransform(unittest.TestCase):
    def setUp(self):
        self.input_img_np = np.expand_dims(TEST_INPUT, axis=0)  # mock image (numpy array)
        self.input_mask_np = np.expand_dims(TEST_MASK, axis=0)  # mock mask (numpy array)

        self.input_img_torch = torch.from_numpy(TEST_INPUT).unsqueeze(0)  # mock image (torch tensor)
        self.input_mask_torch = torch.from_numpy(TEST_MASK).unsqueeze(0)  # mock mask (torch tensor)

    def test_parameters(self):
        # Unknown mode
        with self.assertRaises(ValueError):
            UltrasoundConfidenceMapTransform(mode="unknown")

        # Unknown sink_mode
        with self.assertRaises(ValueError):
            UltrasoundConfidenceMapTransform(sink_mode="unknown")

    def test_rgb(self):
        # RGB image
        input_img_rgb = np.expand_dims(np.repeat(self.input_img_np, 3, axis=0), axis=0)
        input_img_rgb_torch = torch.from_numpy(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="all")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_ALL_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_ALL_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mid")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MID_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MID_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="min")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MIN_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MIN_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mask")
        result_torch = transform(input_img_rgb_torch, self.input_mask_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MASK_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb, self.input_mask_np)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MASK_OUTPUT, rtol=1e-4, atol=1e-4)

    def test_multi_channel_2d(self):
        # 2D multi-channel image
        input_img_rgb = np.expand_dims(np.repeat(self.input_img_np, 17, axis=0), axis=0)
        input_img_rgb_torch = torch.from_numpy(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="all")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_ALL_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_ALL_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mid")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MID_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MID_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="min")
        result_torch = transform(input_img_rgb_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MIN_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MIN_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mask")
        result_torch = transform(input_img_rgb_torch, self.input_mask_torch)
        self.assertIsInstance(result_torch, torch.Tensor)
        assert_allclose(result_torch, torch.tensor(SINK_MASK_OUTPUT), rtol=1e-4, atol=1e-4)
        result_np = transform(input_img_rgb, self.input_mask_np)
        self.assertIsInstance(result_np, np.ndarray)
        assert_allclose(result_np, SINK_MASK_OUTPUT, rtol=1e-4, atol=1e-4)

    def test_non_one_first_dim(self):
        # Image without first dimension as 1
        input_img_rgb = np.repeat(self.input_img_np, 3, axis=0)
        input_img_rgb_torch = torch.from_numpy(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="all")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mid")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="min")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mask")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch, self.input_mask_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb, self.input_mask_np)

    def test_no_first_dim(self):
        # Image without first dimension
        input_img_rgb = self.input_img_np[0]
        input_img_rgb_torch = torch.from_numpy(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="all")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mid")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="min")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb)

        transform = UltrasoundConfidenceMapTransform(sink_mode="mask")
        with self.assertRaises(ValueError):
            transform(input_img_rgb_torch, self.input_mask_torch)
        with self.assertRaises(ValueError):
            transform(input_img_rgb, self.input_mask_np)

    def test_sink_all(self):
        transform = UltrasoundConfidenceMapTransform(sink_mode="all")

        # This should not raise an exception for torch tensor
        result_torch = transform(self.input_img_torch)
        self.assertIsInstance(result_torch, torch.Tensor)

        # This should not raise an exception for numpy array
        result_np = transform(self.input_img_np)
        self.assertIsInstance(result_np, np.ndarray)

    def test_sink_mid(self):
        transform = UltrasoundConfidenceMapTransform(sink_mode="mid")

        # This should not raise an exception for torch tensor
        result_torch = transform(self.input_img_torch)
        self.assertIsInstance(result_torch, torch.Tensor)

        # This should not raise an exception for numpy array
        result_np = transform(self.input_img_np)
        self.assertIsInstance(result_np, np.ndarray)

    def test_sink_min(self):
        transform = UltrasoundConfidenceMapTransform(sink_mode="min")

        # This should not raise an exception for torch tensor
        result_torch = transform(self.input_img_torch)
        self.assertIsInstance(result_torch, torch.Tensor)

        # This should not raise an exception for numpy array
        result_np = transform(self.input_img_np)
        self.assertIsInstance(result_np, np.ndarray)

    def test_sink_mask(self):
        transform = UltrasoundConfidenceMapTransform(sink_mode="mask")

        # This should not raise an exception for torch tensor with mask
        result_torch = transform(self.input_img_torch, self.input_mask_torch)
        self.assertIsInstance(result_torch, torch.Tensor)

        # This should not raise an exception for numpy array with mask
        result_np = transform(self.input_img_np, self.input_mask_np)
        self.assertIsInstance(result_np, np.ndarray)

        # This should raise an exception for torch tensor without mask
        with self.assertRaises(ValueError):
            transform(self.input_img_torch)

        # This should raise an exception for numpy array without mask
        with self.assertRaises(ValueError):
            transform(self.input_img_np)

    def test_func(self):
        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="all")
        output = transform(self.input_img_np)
        assert_allclose(output, SINK_ALL_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="mid")
        output = transform(self.input_img_np)
        assert_allclose(output, SINK_MID_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="min")
        output = transform(self.input_img_np)
        assert_allclose(output, SINK_MIN_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="mask")
        output = transform(self.input_img_np, self.input_mask_np)
        assert_allclose(output, SINK_MASK_OUTPUT, rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="all")
        output = transform(self.input_img_torch)
        assert_allclose(output, torch.tensor(SINK_ALL_OUTPUT), rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="mid")
        output = transform(self.input_img_torch)
        assert_allclose(output, torch.tensor(SINK_MID_OUTPUT), rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="min")
        output = transform(self.input_img_torch)
        assert_allclose(output, torch.tensor(SINK_MIN_OUTPUT), rtol=1e-4, atol=1e-4)

        transform = UltrasoundConfidenceMapTransform(alpha=2.0, beta=90.0, gamma=0.05, mode="B", sink_mode="mask")
        output = transform(self.input_img_torch, self.input_mask_torch)
        assert_allclose(output, torch.tensor(SINK_MASK_OUTPUT), rtol=1e-4, atol=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import unittest
from unittest import skipUnless

import torch
from parameterized import parameterized

from monai.data import CuCIMWSIReader, ImageReader, OpenSlideWSIReader, WSIReader
from monai.inferers import WSISlidingWindowSplitter
from tests.utils import download_url_or_skip_test, optional_import, testing_data_config

cucim, has_cucim = optional_import("cucim")
has_cucim = has_cucim and hasattr(cucim, "CuImage")
_, has_osl = optional_import("openslide")

WSI_READER_STR = None
WSI_READER_CLASS: type[CuCIMWSIReader] | type[OpenSlideWSIReader] | None = None
if has_cucim:
    WSI_READER_STR = "cuCIM"
    WSI_READER_CLASS = CuCIMWSIReader
elif has_osl:
    WSI_READER_STR = "OpenSlide"
    WSI_READER_CLASS = OpenSlideWSIReader

WSI_GENERIC_TIFF_KEY = "wsi_generic_tiff"
WSI_GENERIC_TIFF_PATH = os.path.join(os.path.dirname(__file__), "testing_data", f"temp_{WSI_GENERIC_TIFF_KEY}.tiff")

HEIGHT = 32914
WIDTH = 46000

# ----------------------------------------------------------------------------
# WSI test cases
# ----------------------------------------------------------------------------

TEST_CASE_WSI_0_BASE = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]

TEST_CASE_WSI_1_BASE = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "level": 1, "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]

# Check readers
if WSI_READER_STR is not None:
    TEST_CASE_WSI_2_READER = [
        WSI_GENERIC_TIFF_PATH,
        {"patch_size": (1000, 1000), "reader": WSIReader(backend=WSI_READER_STR)},
        {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
    ]
else:
    TEST_CASE_WSI_2_READER = []
TEST_CASE_WSI_3_READER = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.0, "reader": WSIReader, "backend": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]
TEST_CASE_WSI_4_READER = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.0, "reader": WSI_READER_CLASS},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]
TEST_CASE_WSI_5_READER = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.0, "level": 1, "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]

# Check overlaps
TEST_CASE_WSI_6_OVERLAP = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.0, "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1000), (0, 2000), (0, 3000)]},
]
TEST_CASE_WSI_7_OVERLAP = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.5, "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 500), (0, 1000), (0, 1500)]},
]
TEST_CASE_WSI_8_OVERLAP = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "overlap": 0.999, "reader": WSI_READER_STR},
    {"locations": [(0, 0), (0, 1), (0, 2), (0, 3)]},
]


# Filtering functions test cases
def gen_location_filter(locations):
    def my_filter(patch, loc):
        if loc in locations:
            return False
        return True

    return my_filter


TEST_CASE_WSI_9_FILTER = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (1000, 1000), "reader": WSI_READER_STR, "filter_fn": gen_location_filter([(0, 0), (0, 2000)])},
    {"locations": [(0, 1000), (0, 3000)]},
]


# ----------------------------------------------------------------------------
# Error test cases
# ----------------------------------------------------------------------------
def extra_parameter_filter(patch, location, extra):
    return


def missing_parameter_filter(patch):
    return


# invalid overlap: float 1.0
TEST_CASE_ERROR_0 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "overlap": 1.0, "reader": WSI_READER_STR},
    ValueError,
]
# invalid overlap: negative float
TEST_CASE_ERROR_1 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "overlap": -0.1, "reader": WSI_READER_STR},
    ValueError,
]
# invalid overlap: negative integer
TEST_CASE_ERROR_2 = [WSI_GENERIC_TIFF_PATH, {"patch_size": (2, 2), "overlap": -1, "reader": WSI_READER_STR}, ValueError]
# invalid overlap: integer larger than patch size
TEST_CASE_ERROR_3 = [WSI_GENERIC_TIFF_PATH, {"patch_size": (2, 2), "overlap": 3, "reader": WSI_READER_STR}, ValueError]

# invalid offset: positive and larger than image size
TEST_CASE_ERROR_4 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "offset": WIDTH, "reader": WSI_READER_STR},
    ValueError,
]
# invalid offset: negative and larger than patch size (in magnitude)
TEST_CASE_ERROR_5 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "offset": -3, "pad_mode": "constant", "reader": WSI_READER_STR},
    ValueError,
]
# invalid offset: negative and no padding
TEST_CASE_ERROR_6 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "pad_mode": None, "offset": -1, "reader": WSI_READER_STR},
    ValueError,
]

# invalid filter function: with more than two positional parameters
TEST_CASE_ERROR_7 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "filter_fn": extra_parameter_filter, "reader": WSI_READER_STR},
    ValueError,
]
# invalid filter function: with less than two positional parameters
TEST_CASE_ERROR_8 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "filter_fn": missing_parameter_filter, "reader": WSI_READER_STR},
    ValueError,
]
# invalid filter function: non-callable
TEST_CASE_ERROR_9 = [
    WSI_GENERIC_TIFF_PATH,
    {"patch_size": (2, 2), "filter_fn": 1, "reader": WSI_READER_STR},
    ValueError,
]

# invalid reader
TEST_CASE_ERROR_10 = [WSI_GENERIC_TIFF_PATH, {"patch_size": (2, 2), "reader": ImageReader}, ValueError]


@skipUnless(WSI_READER_STR, "Requires cucim or openslide!")
def setUpModule():
    download_url_or_skip_test(
        testing_data_config("images", WSI_GENERIC_TIFF_KEY, "url"),
        WSI_GENERIC_TIFF_PATH,
        hash_type=testing_data_config("images", WSI_GENERIC_TIFF_KEY, "hash_type"),
        hash_val=testing_data_config("images", WSI_GENERIC_TIFF_KEY, "hash_val"),
    )


class WSISlidingWindowSplitterTests(unittest.TestCase):
    @parameterized.expand(
        [
            TEST_CASE_WSI_0_BASE,
            TEST_CASE_WSI_1_BASE,
            TEST_CASE_WSI_2_READER,
            TEST_CASE_WSI_3_READER,
            TEST_CASE_WSI_4_READER,
            TEST_CASE_WSI_5_READER,
            TEST_CASE_WSI_6_OVERLAP,
            TEST_CASE_WSI_7_OVERLAP,
            TEST_CASE_WSI_8_OVERLAP,
            TEST_CASE_WSI_9_FILTER,
        ]
    )
    def test_split_patches_wsi(self, filepath, arguments, expected):
        patches = WSISlidingWindowSplitter(**arguments)(filepath)
        for sample, expected_loc in zip(patches, expected["locations"]):
            patch = sample[0]
            loc = sample[1]
            self.assertTrue(isinstance(patch, torch.Tensor))
            self.assertTupleEqual(patch.shape[2:], arguments["patch_size"])
            self.assertTrue(isinstance(loc, tuple))
            self.assertTupleEqual(loc, expected_loc)

    @parameterized.expand(
        [
            TEST_CASE_ERROR_0,
            TEST_CASE_ERROR_1,
            TEST_CASE_ERROR_2,
            TEST_CASE_ERROR_3,
            TEST_CASE_ERROR_4,
            TEST_CASE_ERROR_5,
            TEST_CASE_ERROR_6,
            TEST_CASE_ERROR_7,
            TEST_CASE_ERROR_8,
            TEST_CASE_ERROR_9,
            TEST_CASE_ERROR_10,
        ]
    )
    def test_split_patches_errors(self, image, arguments, expected_error):
        with self.assertRaises(expected_error):
            patches = WSISlidingWindowSplitter(**arguments)(image)
            patches = list(patches)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from monai.losses import GeneralizedWassersteinDiceLoss
from tests.utils import test_script_save


class TestGeneralizedWassersteinDiceLoss(unittest.TestCase):
    def test_bin_seg_2d(self):
        target = torch.tensor([[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]])

        # add another dimension corresponding to the batch (batch size = 1 here)
        target = target.unsqueeze(0)
        pred_very_good = 1000 * F.one_hot(target, num_classes=2).permute(0, 3, 1, 2).float()
        pred_very_poor = 1000 * F.one_hot(1 - target, num_classes=2).permute(0, 3, 1, 2).float()

        for weight_mode in ["default", "GDL"]:
            # initialize the loss
            loss = GeneralizedWassersteinDiceLoss(
                dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode=weight_mode
            )

            # the loss for pred_very_good should be close to 0
            loss_good = float(loss.forward(pred_very_good, target))
            self.assertAlmostEqual(loss_good, 0.0, places=3)

            # same test, but with target with a class dimension
            target_4dim = target.unsqueeze(1)
            loss_good = float(loss.forward(pred_very_good, target_4dim))
            self.assertAlmostEqual(loss_good, 0.0, places=3)

            # the loss for pred_very_poor should be close to 1
            loss_poor = float(loss.forward(pred_very_poor, target))
            self.assertAlmostEqual(loss_poor, 1.0, places=3)

    def test_different_target_data_type(self):
        """
        Test if the loss is compatible with all the integer types
        for the target segmentation.
        """
        # define 2d examples
        target = torch.tensor([[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]])

        # add another dimension corresponding to the batch (batch size = 1 here)
        target = target.unsqueeze(0)  # shape (1, H, W)
        pred_very_good = 1000 * F.one_hot(target, num_classes=2).permute(0, 3, 1, 2).float()

        target_uint8 = target.to(torch.uint8)
        target_int8 = target.to(torch.int8)
        target_short = target.short()
        target_int = target.int()
        target_long = target.long()
        target_list = [target_uint8, target_int8, target_short, target_int, target_long]

        for w_mode in ["default", "GDL"]:
            # initialize the loss
            loss = GeneralizedWassersteinDiceLoss(dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode=w_mode)

            # The test should pass irrespectively of the integer type used
            for t in target_list:
                # the loss for pred_very_good should be close to 0
                loss_good = float(loss.forward(pred_very_good, t))
                self.assertAlmostEqual(loss_good, 0.0, places=3)

    def test_empty_class_2d(self):
        num_classes = 2
        target = torch.tensor([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])

        # add another dimension corresponding to the batch (batch size = 1 here)
        target = target.unsqueeze(0)
        pred_very_good = 1000 * F.one_hot(target, num_classes=num_classes).permute(0, 3, 1, 2).float()
        pred_very_poor = 1000 * F.one_hot(1 - target, num_classes=num_classes).permute(0, 3, 1, 2).float()

        for w_mode in ["default", "GDL"]:
            # initialize the loss
            loss = GeneralizedWassersteinDiceLoss(dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode=w_mode)

            # loss for pred_very_good should be close to 0
            loss_good = float(loss.forward(pred_very_good, target))
            self.assertAlmostEqual(loss_good, 0.0, places=3)

            # loss for pred_very_poor should be close to 1
            loss_poor = float(loss.forward(pred_very_poor, target))
            self.assertAlmostEqual(loss_poor, 1.0, places=3)

    def test_bin_seg_3d(self):
        # define 3d examples
        target = torch.tensor(
            [
                # raw 0
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
                # raw 1
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
                # raw 2
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
            ]
        )

        # add another dimension corresponding to the batch (batch size = 1 here)
        target = target.unsqueeze(0)  # shape (1, H, W, D)
        pred_very_good = 1000 * F.one_hot(target, num_classes=2).permute(0, 4, 1, 2, 3).float()
        pred_very_poor = 1000 * F.one_hot(1 - target, num_classes=2).permute(0, 4, 1, 2, 3).float()

        for w_mode in ["default", "GDL"]:
            # initialize the loss
            loss = GeneralizedWassersteinDiceLoss(dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode=w_mode)

            # mean dice loss for pred_very_good should be close to 0
            loss_good = float(loss.forward(pred_very_good, target))
            self.assertAlmostEqual(loss_good, 0.0, places=3)

            # mean dice loss for pred_very_poor should be close to 1
            loss_poor = float(loss.forward(pred_very_poor, target))
            self.assertAlmostEqual(loss_poor, 1.0, places=3)

    def test_convergence(self):
        """
        The goal of this test is to assess if the gradient of the loss function
        is correct by testing if we can train a one layer neural network
        to segment one image.
        We verify that the loss is decreasing in almost all SGD steps.
        """
        learning_rate = 0.001
        max_iter = 50

        # define a simple 3d example
        target_seg = torch.tensor(
            [
                # raw 0
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
                # raw 1
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
                # raw 2
                [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
            ]
        )
        target_seg = torch.unsqueeze(target_seg, dim=0)
        image = 12 * target_seg + 27  # dummy image to segment
        image = image.float()
        num_classes = 2
        num_voxels = 3 * 4 * 4

        # define a model with one layer
        class OnelayerNet(nn.Module):
            def __init__(self):
                super().__init__()
                self.layer = nn.Linear(num_voxels, num_voxels * num_classes)

            def forward(self, x):
                x = x.view(-1, num_voxels)
                x = self.layer(x)
                x = x.view(-1, num_classes, 3, 4, 4)
                return x

        for w_mode in ["default", "GDL"]:
            # initialise the network
            net = OnelayerNet()

            # initialize the loss
            loss = GeneralizedWassersteinDiceLoss(dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode=w_mode)

            # initialize an optimizer
            optimizer = optim.Adam(net.parameters(), lr=learning_rate)

            # initial difference between pred and target
            pred_start = torch.argmax(net(image), dim=1)
            diff_start = torch.norm(pred_start.float() - target_seg.float())

            loss_history = []
            # train the network
            for _ in range(max_iter):
                # set the gradient to zero
                optimizer.zero_grad()

                # forward pass
                output = net(image)
                loss_val = loss(output, target_seg)

                # backward pass
                loss_val.backward()
                optimizer.step()

                # stats
                loss_history.append(loss_val.item())

            # difference between pred and target after training
            pred_end = torch.argmax(net(image), dim=1)
            diff_end = torch.norm(pred_end.float() - target_seg.float())

            # count the number of SGD steps in which the loss decreases
            num_decreasing_steps = 0
            for i in range(len(loss_history) - 1):
                if loss_history[i] > loss_history[i + 1]:
                    num_decreasing_steps += 1
            decreasing_steps_ratio = float(num_decreasing_steps) / (len(loss_history) - 1)

            # verify that the loss is decreasing for sufficiently many SGD steps
            self.assertTrue(decreasing_steps_ratio > 0.75)

            # check that the predicted segmentation has improved
            self.assertGreater(diff_start, diff_end)

    def test_script(self):
        target = torch.tensor([[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]])

        # add another dimension corresponding to the batch (batch size = 1 here)
        target = target.unsqueeze(0)
        pred_very_good = 1000 * F.one_hot(target, num_classes=2).permute(0, 3, 1, 2).float()

        loss = GeneralizedWassersteinDiceLoss(dist_matrix=np.array([[0.0, 1.0], [1.0, 0.0]]), weighting_mode="default")

        test_script_save(loss, pred_very_good, target)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.data.meta_obj import set_track_meta
from monai.data.meta_tensor import MetaTensor
from monai.transforms import Flip
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.utils import TEST_DEVICES, TEST_NDARRAYS_ALL, NumpyImageTestCase2D, assert_allclose, test_local_inversion

INVALID_CASES = [("wrong_axis", ["s", 1], TypeError), ("not_numbers", "s", TypeError)]

VALID_CASES = [("no_axis", None), ("one_axis", 1), ("many_axis", [0, 1]), ("negative_axis", [0, -1])]

TORCH_CASES = []
for track_meta in (False, True):
    for device in TEST_DEVICES:
        TORCH_CASES.append([[0, 1], torch.zeros((1, 3, 2)), track_meta, *device])


class TestFlip(NumpyImageTestCase2D):
    @parameterized.expand(INVALID_CASES)
    def test_invalid_inputs(self, _, spatial_axis, raises):
        with self.assertRaises(raises):
            flip = Flip(spatial_axis)
            flip(self.imt[0])

    @parameterized.expand(VALID_CASES)
    def test_correct_results(self, _, spatial_axis):
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            init_param = {"spatial_axis": spatial_axis}
            flip = Flip(**init_param)
            expected = [np.flip(channel, spatial_axis) for channel in self.imt[0]]
            expected = np.stack(expected)
            call_param = {"img": im}
            result = flip(**call_param)
            test_resampler_lazy(flip, result, init_param, call_param)
            assert_allclose(result, p(expected), type_test="tensor")
            test_local_inversion(flip, result, im)

    @parameterized.expand(TORCH_CASES)
    def test_torch(self, spatial_axis, img: torch.Tensor, track_meta: bool, device):
        set_track_meta(track_meta)
        img = img.to(device)
        init_param = {"spatial_axis": spatial_axis}
        xform = Flip(**init_param)
        call_param = {"img": img}
        res = xform(**call_param)  # type: ignore[arg-type]
        self.assertEqual(img.shape, res.shape)
        if track_meta:
            test_resampler_lazy(xform, res, init_param, call_param)
            self.assertIsInstance(res, MetaTensor)
        else:
            self.assertNotIsInstance(res, MetaTensor)
            self.assertIsInstance(res, torch.Tensor)
            with self.assertRaisesRegex(ValueError, "MetaTensor"):
                xform.inverse(res)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A collection of generic interfaces for MONAI transforms.
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Hashable, Iterable, Mapping
from typing import Any, TypeVar

import numpy as np
import torch

from monai import config, transforms
from monai.config import KeysCollection
from monai.data.meta_tensor import MetaTensor
from monai.transforms.traits import LazyTrait, RandomizableTrait, ThreadUnsafe
from monai.utils import MAX_SEED, ensure_tuple, first
from monai.utils.enums import TransformBackends
from monai.utils.misc import MONAIEnvVars

__all__ = [
    "ThreadUnsafe",
    "apply_transform",
    "Randomizable",
    "LazyTransform",
    "RandomizableTransform",
    "Transform",
    "MapTransform",
]

ReturnType = TypeVar("ReturnType")


def _apply_transform(
    transform: Callable[..., ReturnType],
    data: Any,
    unpack_parameters: bool = False,
    lazy: bool | None = False,
    overrides: dict | None = None,
    logger_name: bool | str = False,
) -> ReturnType:
    """
    Perform a transform 'transform' on 'data', according to the other parameters specified.

    If `data` is a tuple and `unpack_parameters` is True, each parameter of `data` is unpacked
    as arguments to `transform`. Otherwise `data` is considered as single argument to `transform`.

    If 'lazy' is True, this method first checks whether it can execute this method lazily. If it
    can't, it will ensure that all pending lazy transforms on 'data' are applied before applying
    this 'transform' to it. If 'lazy' is True, and 'overrides' are provided, those overrides will
    be applied to the pending operations on 'data'. See ``Compose`` for more details on lazy
    resampling, which is an experimental feature for 1.2.

    Please note, this class is function is designed to be called by ``apply_transform``.
    In general, you should not need to make specific use of it unless you are implementing
    pipeline execution mechanisms.

    Args:
        transform: a callable to be used to transform `data`.
        data: the tensorlike or dictionary of tensorlikes to be executed on
        unpack_parameters: whether to unpack parameters for `transform`. Defaults to False.
        lazy: whether to enable lazy evaluation for lazy transforms. If False, transforms will be
            carried out on a transform by transform basis. If True, all lazy transforms will
            be executed by accumulating changes and resampling as few times as possible.
            See the :ref:`Lazy Resampling topic<lazy_resampling> for more information about lazy resampling.
        overrides: this optional parameter allows you to specify a dictionary of parameters that should be overridden
            when executing a pipeline. These each parameter that is compatible with a given transform is then applied
            to that transform before it is executed. Note that overrides are currently only applied when
            :ref:`Lazy Resampling<lazy_resampling>` is enabled for the pipeline or a given transform. If lazy is False
            they are ignored. Currently supported args are:
            {``"mode"``, ``"padding_mode"``, ``"dtype"``, ``"align_corners"``, ``"resample_mode"``, ``device``}.
        logger_name: this optional parameter allows you to specify a logger by name for logging of pipeline execution.
            Setting this to False disables logging. Setting it to True enables logging to the default loggers.
            Setting a string overrides the logger name to which logging is performed.

    Returns:
        ReturnType: The return type of `transform`.
    """
    from monai.transforms.lazy.functional import apply_pending_transforms_in_order

    data = apply_pending_transforms_in_order(transform, data, lazy, overrides, logger_name)

    if isinstance(data, tuple) and unpack_parameters:
        return transform(*data, lazy=lazy) if isinstance(transform, LazyTrait) else transform(*data)

    return transform(data, lazy=lazy) if isinstance(transform, LazyTrait) else transform(data)


def apply_transform(
    transform: Callable[..., ReturnType],
    data: Any,
    map_items: bool = True,
    unpack_items: bool = False,
    log_stats: bool | str = False,
    lazy: bool | None = None,
    overrides: dict | None = None,
) -> list[ReturnType] | ReturnType:
    """
    Transform `data` with `transform`.

    If `data` is a list or tuple and `map_data` is True, each item of `data` will be transformed
    and this method returns a list of outcomes.
    otherwise transform will be applied once with `data` as the argument.

    Args:
        transform: a callable to be used to transform `data`.
        data: an object to be transformed.
        map_items: whether to apply transform to each item in `data`,
            if `data` is a list or tuple. Defaults to True.
        unpack_items: whether to unpack parameters using `*`. Defaults to False.
        log_stats: log errors when they occur in the processing pipeline. By default, this is set to False, which
            disables the logger for processing pipeline errors. Setting it to None or True will enable logging to the
            default logger name. Setting it to a string specifies the logger to which errors should be logged.
        lazy: whether to execute in lazy mode or not. See the :ref:`Lazy Resampling topic<lazy_resampling> for more
            information about lazy resampling. Defaults to None.
        overrides: optional overrides to apply to transform parameters. This parameter is ignored unless transforms
            are being executed lazily. See the :ref:`Lazy Resampling topic<lazy_resampling> for more details and
            examples of its usage.

    Raises:
        Exception: When ``transform`` raises an exception.

    Returns:
        Union[List[ReturnType], ReturnType]: The return type of `transform` or a list thereof.
    """
    try:
        if isinstance(data, (list, tuple)) and map_items:
            return [_apply_transform(transform, item, unpack_items, lazy, overrides, log_stats) for item in data]
        return _apply_transform(transform, data, unpack_items, lazy, overrides, log_stats)
    except Exception as e:
        # if in debug mode, don't swallow exception so that the breakpoint
        # appears where the exception was raised.
        if MONAIEnvVars.debug():
            raise
        if log_stats is not False and not isinstance(transform, transforms.compose.Compose):
            # log the input data information of exact transform in the transform chain
            if isinstance(log_stats, str):
                datastats = transforms.utility.array.DataStats(data_shape=False, value_range=False, name=log_stats)
            else:
                datastats = transforms.utility.array.DataStats(data_shape=False, value_range=False)
            logger = logging.getLogger(datastats._logger_name)
            logger.error(f"\n=== Transform input info -- {type(transform).__name__} ===")
            if isinstance(data, (list, tuple)):
                data = data[0]

            def _log_stats(data, prefix: str | None = "Data"):
                if isinstance(data, (np.ndarray, torch.Tensor)):
                    # log data type, shape, range for array
                    datastats(img=data, data_shape=True, value_range=True, prefix=prefix)
                else:
                    # log data type and value for other metadata
                    datastats(img=data, data_value=True, prefix=prefix)

            if isinstance(data, dict):
                for k, v in data.items():
                    _log_stats(data=v, prefix=k)
            else:
                _log_stats(data=data)
        raise RuntimeError(f"applying transform {transform}") from e


class Randomizable(ThreadUnsafe, RandomizableTrait):
    """
    An interface for handling random state locally, currently based on a class
    variable `R`, which is an instance of `np.random.RandomState`.  This
    provides the flexibility of component-specific determinism without
    affecting the global states.  It is recommended to use this API with
    :py:class:`monai.data.DataLoader` for deterministic behaviour of the
    preprocessing pipelines. This API is not thread-safe. Additionally,
    deepcopying instance of this class often causes insufficient randomness as
    the random states will be duplicated.
    """

    R: np.random.RandomState = np.random.RandomState()

    def set_random_state(self, seed: int | None = None, state: np.random.RandomState | None = None) -> Randomizable:
        """
        Set the random state locally, to control the randomness, the derived
        classes should use :py:attr:`self.R` instead of `np.random` to introduce random
        factors.

        Args:
            seed: set the random state with an integer seed.
            state: set the random state with a `np.random.RandomState` object.

        Raises:
            TypeError: When ``state`` is not an ``Optional[np.random.RandomState]``.

        Returns:
            a Randomizable instance.

        """
        if seed is not None:
            _seed = id(seed) if not isinstance(seed, (int, np.integer)) else seed
            _seed = _seed % MAX_SEED
            self.R = np.random.RandomState(_seed)
            return self

        if state is not None:
            if not isinstance(state, np.random.RandomState):
                raise TypeError(f"state must be None or a np.random.RandomState but is {type(state).__name__}.")
            self.R = state
            return self

        self.R = np.random.RandomState()
        return self

    def randomize(self, data: Any) -> None:
        """
        Within this method, :py:attr:`self.R` should be used, instead of `np.random`, to introduce random factors.

        all :py:attr:`self.R` calls happen here so that we have a better chance to
        identify errors of sync the random state.

        This method can generate the random factors based on properties of the input data.

        Raises:
            NotImplementedError: When the subclass does not override this method.
        """
        raise NotImplementedError(f"Subclass {self.__class__.__name__} must implement this method.")


class Transform(ABC):
    """
    An abstract class of a ``Transform``.
    A transform is callable that processes ``data``.

    It could be stateful and may modify ``data`` in place,
    the implementation should be aware of:

        #. thread safety when mutating its own states.
           When used from a multi-process context, transform's instance variables are read-only.
           thread-unsafe transforms should inherit :py:class:`monai.transforms.ThreadUnsafe`.
        #. ``data`` content unused by this transform may still be used in the
           subsequent transforms in a composed transform.
        #. storing too much information in ``data`` may cause some memory issue or IPC sync issue,
           especially in the multi-processing environment of PyTorch DataLoader.

    See Also

        :py:class:`monai.transforms.Compose`
    """

    # Transforms should add `monai.transforms.utils.TransformBackends` to this list if they are performing
    # the data processing using the corresponding backend APIs.
    # Most of MONAI transform's inputs and outputs will be converted into torch.Tensor or monai.data.MetaTensor.
    # This variable provides information about whether the input will be converted
    # to other data types during the transformation. Note that not all `dtype` (such as float32, uint8) are supported
    # by all the data types, the `dtype` during the conversion is determined automatically by each transform,
    # please refer to the transform's docstring.
    backend: list[TransformBackends] = []

    @abstractmethod
    def __call__(self, data: Any):
        """
        ``data`` is an element which often comes from an iteration over an
        iterable, such as :py:class:`torch.utils.data.Dataset`. This method should
        return an updated version of ``data``.
        To simplify the input validations, most of the transforms assume that

        - ``data`` is a Numpy ndarray, PyTorch Tensor or string,
        - the data shape can be:

          #. string data without shape, `LoadImage` transform expects file paths,
          #. most of the pre-/post-processing transforms expect: ``(num_channels, spatial_dim_1[, spatial_dim_2, ...])``,
             except for example: `AddChannel` expects (spatial_dim_1[, spatial_dim_2, ...])

        - the channel dimension is often not omitted even if number of channels is one.

        This method can optionally take additional arguments to help execute transformation operation.

        Raises:
            NotImplementedError: When the subclass does not override this method.

        """
        raise NotImplementedError(f"Subclass {self.__class__.__name__} must implement this method.")


class LazyTransform(Transform, LazyTrait):
    """
    An implementation of functionality for lazy transforms that can be subclassed by array and
    dictionary transforms to simplify implementation of new lazy transforms.
    """

    def __init__(self, lazy: bool | None = False):
        if lazy is not None:
            if not isinstance(lazy, bool):
                raise TypeError(f"lazy must be a bool but is of type {type(lazy)}")
        self._lazy = lazy

    @property
    def lazy(self):
        return self._lazy

    @lazy.setter
    def lazy(self, lazy: bool | None):
        if lazy is not None:
            if not isinstance(lazy, bool):
                raise TypeError(f"lazy must be a bool but is of type {type(lazy)}")
        self._lazy = lazy

    @property
    def requires_current_data(self):
        return False


class RandomizableTransform(Randomizable, Transform):
    """
    An interface for handling random state locally, currently based on a class variable `R`,
    which is an instance of `np.random.RandomState`.
    This class introduces a randomized flag `_do_transform`, is mainly for randomized data augmentation transforms.
    For example:

    .. code-block:: python

        from monai.transforms import RandomizableTransform

        class RandShiftIntensity100(RandomizableTransform):
            def randomize(self):
                super().randomize(None)
                self._offset = self.R.uniform(low=0, high=100)

            def __call__(self, img):
                self.randomize()
                if not self._do_transform:
                    return img
                return img + self._offset

        transform = RandShiftIntensity()
        transform.set_random_state(seed=0)
        print(transform(10))

    """

    def __init__(self, prob: float = 1.0, do_transform: bool = True):
        self._do_transform = do_transform
        self.prob = min(max(prob, 0.0), 1.0)

    def randomize(self, data: Any) -> None:
        """
        Within this method, :py:attr:`self.R` should be used, instead of `np.random`, to introduce random factors.

        all :py:attr:`self.R` calls happen here so that we have a better chance to
        identify errors of sync the random state.

        This method can generate the random factors based on properties of the input data.
        """
        self._do_transform = self.R.rand() < self.prob


class MapTransform(Transform):
    """
    A subclass of :py:class:`monai.transforms.Transform` with an assumption
    that the ``data`` input of ``self.__call__`` is a MutableMapping such as ``dict``.

    The ``keys`` parameter will be used to get and set the actual data
    item to transform.  That is, the callable of this transform should
    follow the pattern:

        .. code-block:: python

            def __call__(self, data):
                for key in self.keys:
                    if key in data:
                        # update output data with some_transform_function(data[key]).
                    else:
                        # raise exception unless allow_missing_keys==True.
                return data

    Raises:
        ValueError: When ``keys`` is an empty iterable.
        TypeError: When ``keys`` type is not in ``Union[Hashable, Iterable[Hashable]]``.

    """

    def __new__(cls, *args, **kwargs):
        if config.USE_META_DICT:
            # call_update after MapTransform.__call__
            cls.__call__ = transforms.attach_hook(cls.__call__, MapTransform.call_update, "post")  # type: ignore

            if hasattr(cls, "inverse"):
                # inverse_update before InvertibleTransform.inverse
                cls.inverse: Any = transforms.attach_hook(cls.inverse, transforms.InvertibleTransform.inverse_update)
        return Transform.__new__(cls)

    def __init__(self, keys: KeysCollection, allow_missing_keys: bool = False) -> None:
        super().__init__()
        self.keys: tuple[Hashable, ...] = ensure_tuple(keys)
        self.allow_missing_keys = allow_missing_keys
        if not self.keys:
            raise ValueError("keys must be non empty.")
        for key in self.keys:
            if not isinstance(key, Hashable):
                raise TypeError(f"keys must be one of (Hashable, Iterable[Hashable]) but is {type(keys).__name__}.")

    def call_update(self, data):
        """
        This function is to be called after every `self.__call__(data)`,
        update `data[key_transforms]` and `data[key_meta_dict]` using the content from MetaTensor `data[key]`,
        for MetaTensor backward compatibility 0.9.0.
        """
        if not isinstance(data, (list, tuple, Mapping)):
            return data
        is_dict = False
        if isinstance(data, Mapping):
            data, is_dict = [data], True
        if not data or not isinstance(data[0], Mapping):
            return data[0] if is_dict else data
        list_d = [dict(x) for x in data]  # list of dict for crop samples
        for idx, dict_i in enumerate(list_d):
            for k in dict_i:
                if not isinstance(dict_i[k], MetaTensor):
                    continue
                list_d[idx] = transforms.sync_meta_info(k, dict_i, t=not isinstance(self, transforms.InvertD))
        return list_d[0] if is_dict else list_d

    @abstractmethod
    def __call__(self, data):
        """
        ``data`` often comes from an iteration over an iterable,
        such as :py:class:`torch.utils.data.Dataset`.

        To simplify the input validations, this method assumes:

        - ``data`` is a Python dictionary,
        - ``data[key]`` is a Numpy ndarray, PyTorch Tensor or string, where ``key`` is an element
          of ``self.keys``, the data shape can be:

          #. string data without shape, `LoadImaged` transform expects file paths,
          #. most of the pre-/post-processing transforms expect: ``(num_channels, spatial_dim_1[, spatial_dim_2, ...])``,
             except for example: `AddChanneld` expects (spatial_dim_1[, spatial_dim_2, ...])

        - the channel dimension is often not omitted even if number of channels is one.

        Raises:
            NotImplementedError: When the subclass does not override this method.

        returns:
            An updated dictionary version of ``data`` by applying the transform.

        """
        raise NotImplementedError(f"Subclass {self.__class__.__name__} must implement this method.")

    def key_iterator(self, data: Mapping[Hashable, Any], *extra_iterables: Iterable | None) -> Generator:
        """
        Iterate across keys and optionally extra iterables. If key is missing, exception is raised if
        `allow_missing_keys==False` (default). If `allow_missing_keys==True`, key is skipped.

        Args:
            data: data that the transform will be applied to
            extra_iterables: anything else to be iterated through
        """
        # if no extra iterables given, create a dummy list of Nones
        ex_iters = extra_iterables or [[None] * len(self.keys)]

        # loop over keys and any extra iterables
        _ex_iters: list[Any]
        for key, *_ex_iters in zip(self.keys, *ex_iters):
            # all normal, yield (what we yield depends on whether extra iterables were given)
            if key in data:
                yield (key,) + tuple(_ex_iters) if extra_iterables else key
            elif not self.allow_missing_keys:
                raise KeyError(
                    f"Key `{key}` of transform `{self.__class__.__name__}` was missing in the data"
                    " and allow_missing_keys==False."
                )

    def first_key(self, data: dict[Hashable, Any]):
        """
        Get the first available key of `self.keys` in the input `data` dictionary.
        If no available key, return an empty tuple `()`.

        Args:
            data: data that the transform will be applied to.

        """
        return first(self.key_iterator(data), ())



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.losses import PerceptualLoss
from monai.utils import optional_import
from tests.utils import SkipIfBeforePyTorchVersion, skip_if_downloading_fails

_, has_torchvision = optional_import("torchvision")
TEST_CASES = [
    [{"spatial_dims": 2, "network_type": "squeeze"}, (2, 1, 64, 64), (2, 1, 64, 64)],
    [
        {"spatial_dims": 3, "network_type": "squeeze", "is_fake_3d": True, "fake_3d_ratio": 0.1},
        (2, 1, 64, 64, 64),
        (2, 1, 64, 64, 64),
    ],
    [{"spatial_dims": 2, "network_type": "radimagenet_resnet50"}, (2, 1, 64, 64), (2, 1, 64, 64)],
    [{"spatial_dims": 2, "network_type": "radimagenet_resnet50"}, (2, 3, 64, 64), (2, 3, 64, 64)],
    [
        {"spatial_dims": 3, "network_type": "radimagenet_resnet50", "is_fake_3d": True, "fake_3d_ratio": 0.1},
        (2, 1, 64, 64, 64),
        (2, 1, 64, 64, 64),
    ],
    [
        {"spatial_dims": 3, "network_type": "medicalnet_resnet10_23datasets", "is_fake_3d": False},
        (2, 1, 64, 64, 64),
        (2, 1, 64, 64, 64),
    ],
    [
        {"spatial_dims": 3, "network_type": "resnet50", "is_fake_3d": True, "pretrained": True, "fake_3d_ratio": 0.2},
        (2, 1, 64, 64, 64),
        (2, 1, 64, 64, 64),
    ],
]


@SkipIfBeforePyTorchVersion((1, 11))
@unittest.skipUnless(has_torchvision, "Requires torchvision")
class TestPerceptualLoss(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_shape(self, input_param, input_shape, target_shape):
        loss = PerceptualLoss(**input_param)
        result = loss(torch.randn(input_shape), torch.randn(target_shape))
        self.assertEqual(result.shape, torch.Size([]))

    @parameterized.expand(TEST_CASES)
    def test_identical_input(self, input_param, input_shape, target_shape):
        with skip_if_downloading_fails():
            loss = PerceptualLoss(**input_param)
        tensor = torch.randn(input_shape)
        result = loss(tensor, tensor)
        self.assertEqual(result, torch.Tensor([0.0]))

    def test_different_shape(self):
        with skip_if_downloading_fails():
            loss = PerceptualLoss(spatial_dims=2, network_type="squeeze")
        tensor = torch.randn(2, 1, 64, 64)
        target = torch.randn(2, 1, 32, 32)
        with self.assertRaises(ValueError):
            loss(tensor, target)

    def test_1d(self):
        with self.assertRaises(NotImplementedError):
            PerceptualLoss(spatial_dims=1)

    def test_medicalnet_on_2d_data(self):
        with self.assertRaises(ValueError):
            PerceptualLoss(spatial_dims=2, network_type="medicalnet_resnet10_23datasets")

        with self.assertRaises(ValueError):
            PerceptualLoss(spatial_dims=2, network_type="medicalnet_resnet50_23datasets")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defines factories for creating layers in generic, extensible, and dimensionally independent ways. A separate factory
object is created for each type of layer, and factory functions keyed to names are added to these objects. Whenever
a layer is requested the factory name and any necessary arguments are passed to the factory object. The return value
is typically a type but can be any callable producing a layer object.

The factory objects contain functions keyed to names converted to upper case, these names can be referred to as members
of the factory so that they can function as constant identifiers. eg. instance normalization is named `Norm.INSTANCE`.

For example, to get a transpose convolution layer the name is needed and then a dimension argument is provided which is
passed to the factory function:

.. code-block:: python

    dimension = 3
    name = Conv.CONVTRANS
    conv = Conv[name, dimension]

This allows the `dimension` value to be set in the constructor, for example so that the dimensionality of a network is
parameterizable. Not all factories require arguments after the name, the caller must be aware which are required.

Defining new factories involves creating the object then associating it with factory functions:

.. code-block:: python

    fact = LayerFactory()

    @fact.factory_function('test')
    def make_something(x, y):
        # do something with x and y to choose which layer type to return
        return SomeLayerType
    ...

    # request object from factory TEST with 1 and 2 as values for x and y
    layer = fact[fact.TEST, 1, 2]

Typically the caller of a factory would know what arguments to pass (ie. the dimensionality of the requested type) but
can be parameterized with the factory name and the arguments to pass to the created type at instantiation time:

.. code-block:: python

    def use_factory(fact_args):
        fact_name, type_args = split_args
        layer_type = fact[fact_name, 1, 2]
        return layer_type(**type_args)
    ...

    kw_args = {'arg0':0, 'arg1':True}
    layer = use_factory( (fact.TEST, kwargs) )
"""

from __future__ import annotations

import warnings
from collections.abc import Callable
from typing import Any

import torch.nn as nn

from monai.networks.utils import has_nvfuser_instance_norm
from monai.utils import look_up_option, optional_import

__all__ = ["LayerFactory", "Dropout", "Norm", "Act", "Conv", "Pool", "Pad", "split_args"]


class LayerFactory:
    """
    Factory object for creating layers, this uses given factory functions to actually produce the types or constructing
    callables. These functions are referred to by name and can be added at any time.
    """

    def __init__(self) -> None:
        self.factories: dict[str, Callable] = {}

    @property
    def names(self) -> tuple[str, ...]:
        """
        Produces all factory names.
        """

        return tuple(self.factories)

    def add_factory_callable(self, name: str, func: Callable) -> None:
        """
        Add the factory function to this object under the given name.
        """

        self.factories[name.upper()] = func
        self.__doc__ = (
            "The supported member"
            + ("s are: " if len(self.names) > 1 else " is: ")
            + ", ".join(f"``{name}``" for name in self.names)
            + ".\nPlease see :py:class:`monai.networks.layers.split_args` for additional args parsing."
        )

    def factory_function(self, name: str) -> Callable:
        """
        Decorator for adding a factory function with the given name.
        """

        def _add(func: Callable) -> Callable:
            self.add_factory_callable(name, func)
            return func

        return _add

    def get_constructor(self, factory_name: str, *args) -> Any:
        """
        Get the constructor for the given factory name and arguments.

        Raises:
            TypeError: When ``factory_name`` is not a ``str``.

        """

        if not isinstance(factory_name, str):
            raise TypeError(f"factory_name must a str but is {type(factory_name).__name__}.")

        func = look_up_option(factory_name.upper(), self.factories)
        return func(*args)

    def __getitem__(self, args) -> Any:
        """
        Get the given name or name/arguments pair. If `args` is a callable it is assumed to be the constructor
        itself and is returned, otherwise it should be the factory name or a pair containing the name and arguments.
        """

        # `args[0]` is actually a type or constructor
        if callable(args):
            return args

        # `args` is a factory name or a name with arguments
        if isinstance(args, str):
            name_obj, args = args, ()
        else:
            name_obj, *args = args

        return self.get_constructor(name_obj, *args)

    def __getattr__(self, key):
        """
        If `key` is a factory name, return it, otherwise behave as inherited. This allows referring to factory names
        as if they were constants, eg. `Fact.FOO` for a factory Fact with factory function foo.
        """

        if key in self.factories:
            return key

        return super().__getattribute__(key)


def split_args(args):
    """
    Split arguments in a way to be suitable for using with the factory types. If `args` is a string it's interpreted as
    the type name.

    Args:
        args (str or a tuple of object name and kwarg dict): input arguments to be parsed.

    Raises:
        TypeError: When ``args`` type is not in ``Union[str, Tuple[Union[str, Callable], dict]]``.

    Examples::

        >>> act_type, args = split_args("PRELU")
        >>> monai.networks.layers.Act[act_type]
        <class 'torch.nn.modules.activation.PReLU'>

        >>> act_type, args = split_args(("PRELU", {"num_parameters": 1, "init": 0.25}))
        >>> monai.networks.layers.Act[act_type](**args)
        PReLU(num_parameters=1)

    """

    if isinstance(args, str):
        return args, {}
    name_obj, name_args = args

    if not (isinstance(name_obj, str) or callable(name_obj)) or not isinstance(name_args, dict):
        msg = "Layer specifiers must be single strings or pairs of the form (name/object-types, argument dict)"
        raise TypeError(msg)

    return name_obj, name_args


# Define factories for these layer types

Dropout = LayerFactory()
Norm = LayerFactory()
Act = LayerFactory()
Conv = LayerFactory()
Pool = LayerFactory()
Pad = LayerFactory()


@Dropout.factory_function("dropout")
def dropout_factory(dim: int) -> type[nn.Dropout | nn.Dropout2d | nn.Dropout3d]:
    types = (nn.Dropout, nn.Dropout2d, nn.Dropout3d)
    return types[dim - 1]


@Dropout.factory_function("alphadropout")
def alpha_dropout_factory(_dim):
    return nn.AlphaDropout


@Norm.factory_function("instance")
def instance_factory(dim: int) -> type[nn.InstanceNorm1d | nn.InstanceNorm2d | nn.InstanceNorm3d]:
    types = (nn.InstanceNorm1d, nn.InstanceNorm2d, nn.InstanceNorm3d)
    return types[dim - 1]


@Norm.factory_function("batch")
def batch_factory(dim: int) -> type[nn.BatchNorm1d | nn.BatchNorm2d | nn.BatchNorm3d]:
    types = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)
    return types[dim - 1]


@Norm.factory_function("group")
def group_factory(_dim) -> type[nn.GroupNorm]:
    return nn.GroupNorm


@Norm.factory_function("layer")
def layer_factory(_dim) -> type[nn.LayerNorm]:
    return nn.LayerNorm


@Norm.factory_function("localresponse")
def local_response_factory(_dim) -> type[nn.LocalResponseNorm]:
    return nn.LocalResponseNorm


@Norm.factory_function("syncbatch")
def sync_batch_factory(_dim) -> type[nn.SyncBatchNorm]:
    return nn.SyncBatchNorm


@Norm.factory_function("instance_nvfuser")
def instance_nvfuser_factory(dim):
    """
    `InstanceNorm3dNVFuser` is a faster version of InstanceNorm layer and implemented in `apex`.
    It only supports 3d tensors as the input. It also requires to use with CUDA and non-Windows OS.
    In this function, if the required library `apex.normalization.InstanceNorm3dNVFuser` does not exist,
    `nn.InstanceNorm3d` will be returned instead.
    This layer is based on a customized autograd function, which is not supported in TorchScript currently.
    Please switch to use `nn.InstanceNorm3d` if TorchScript is necessary.

    Please check the following link for more details about how to install `apex`:
    https://github.com/NVIDIA/apex#installation

    """

    if dim != 3:
        types = (nn.InstanceNorm1d, nn.InstanceNorm2d)
        warnings.warn(f"`InstanceNorm3dNVFuser` only supports 3d cases, use {types[dim - 1]} instead.")
        return types[dim - 1]

    if not has_nvfuser_instance_norm():
        warnings.warn(
            "`apex.normalization.InstanceNorm3dNVFuser` is not installed properly, use nn.InstanceNorm3d instead."
        )
        return nn.InstanceNorm3d
    return optional_import("apex.normalization", name="InstanceNorm3dNVFuser")[0]


Act.add_factory_callable("elu", lambda: nn.modules.ELU)
Act.add_factory_callable("relu", lambda: nn.modules.ReLU)
Act.add_factory_callable("leakyrelu", lambda: nn.modules.LeakyReLU)
Act.add_factory_callable("prelu", lambda: nn.modules.PReLU)
Act.add_factory_callable("relu6", lambda: nn.modules.ReLU6)
Act.add_factory_callable("selu", lambda: nn.modules.SELU)
Act.add_factory_callable("celu", lambda: nn.modules.CELU)
Act.add_factory_callable("gelu", lambda: nn.modules.GELU)
Act.add_factory_callable("sigmoid", lambda: nn.modules.Sigmoid)
Act.add_factory_callable("tanh", lambda: nn.modules.Tanh)
Act.add_factory_callable("softmax", lambda: nn.modules.Softmax)
Act.add_factory_callable("logsoftmax", lambda: nn.modules.LogSoftmax)


@Act.factory_function("swish")
def swish_factory():
    from monai.networks.blocks.activation import Swish

    return Swish


@Act.factory_function("memswish")
def memswish_factory():
    from monai.networks.blocks.activation import MemoryEfficientSwish

    return MemoryEfficientSwish


@Act.factory_function("mish")
def mish_factory():
    from monai.networks.blocks.activation import Mish

    return Mish


@Act.factory_function("geglu")
def geglu_factory():
    from monai.networks.blocks.activation import GEGLU

    return GEGLU


@Conv.factory_function("conv")
def conv_factory(dim: int) -> type[nn.Conv1d | nn.Conv2d | nn.Conv3d]:
    types = (nn.Conv1d, nn.Conv2d, nn.Conv3d)
    return types[dim - 1]


@Conv.factory_function("convtrans")
def convtrans_factory(dim: int) -> type[nn.ConvTranspose1d | nn.ConvTranspose2d | nn.ConvTranspose3d]:
    types = (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)
    return types[dim - 1]


@Pool.factory_function("max")
def maxpooling_factory(dim: int) -> type[nn.MaxPool1d | nn.MaxPool2d | nn.MaxPool3d]:
    types = (nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d)
    return types[dim - 1]


@Pool.factory_function("adaptivemax")
def adaptive_maxpooling_factory(dim: int) -> type[nn.AdaptiveMaxPool1d | nn.AdaptiveMaxPool2d | nn.AdaptiveMaxPool3d]:
    types = (nn.AdaptiveMaxPool1d, nn.AdaptiveMaxPool2d, nn.AdaptiveMaxPool3d)
    return types[dim - 1]


@Pool.factory_function("avg")
def avgpooling_factory(dim: int) -> type[nn.AvgPool1d | nn.AvgPool2d | nn.AvgPool3d]:
    types = (nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d)
    return types[dim - 1]


@Pool.factory_function("adaptiveavg")
def adaptive_avgpooling_factory(dim: int) -> type[nn.AdaptiveAvgPool1d | nn.AdaptiveAvgPool2d | nn.AdaptiveAvgPool3d]:
    types = (nn.AdaptiveAvgPool1d, nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool3d)
    return types[dim - 1]


@Pad.factory_function("replicationpad")
def replication_pad_factory(dim: int) -> type[nn.ReplicationPad1d | nn.ReplicationPad2d | nn.ReplicationPad3d]:
    types = (nn.ReplicationPad1d, nn.ReplicationPad2d, nn.ReplicationPad3d)
    return types[dim - 1]


@Pad.factory_function("constantpad")
def constant_pad_factory(dim: int) -> type[nn.ConstantPad1d | nn.ConstantPad2d | nn.ConstantPad3d]:
    types = (nn.ConstantPad1d, nn.ConstantPad2d, nn.ConstantPad3d)
    return types[dim - 1]



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np

from monai.transforms import IntensityStatsd, RandShiftIntensityd
from monai.utils.enums import PostFix
from tests.utils import TEST_NDARRAYS, NumpyImageTestCase2D, assert_allclose


class TestRandShiftIntensityd(NumpyImageTestCase2D):
    def test_value(self):
        key = "img"
        for p in TEST_NDARRAYS:
            shifter = RandShiftIntensityd(keys=[key], offsets=1.0, prob=1.0)
            shifter.set_random_state(seed=0)
            result = shifter({key: p(self.imt)})
            np.random.seed(0)
            # simulate the randomize() of transform
            np.random.random()
            expected = self.imt + np.random.uniform(low=-1.0, high=1.0)
            assert_allclose(result[key], p(expected), type_test="tensor")

    def test_factor(self):
        key = "img"
        stats = IntensityStatsd(keys=key, ops="max", key_prefix="orig")
        shifter = RandShiftIntensityd(keys=[key], offsets=1.0, factor_key=["orig_max"], prob=1.0)
        data = {key: self.imt, PostFix.meta(key): {"affine": None}}
        shifter.set_random_state(seed=0)
        result = shifter(stats(data))
        np.random.seed(0)
        # simulate the randomize() of transform
        np.random.random()
        expected = self.imt + np.random.uniform(low=-1.0, high=1.0) * np.nanmax(self.imt)
        np.testing.assert_allclose(result[key], expected)

    def test_channel_wise(self):
        key = "img"
        for p in TEST_NDARRAYS:
            scaler = RandShiftIntensityd(keys=[key], offsets=3.0, prob=1.0, channel_wise=True)
            scaler.set_random_state(seed=0)
            result = scaler({key: p(self.imt)})
            np.random.seed(0)
            # simulate the randomize function of transform
            np.random.random()
            channel_num = self.imt.shape[0]
            factor = [np.random.uniform(low=-3.0, high=3.0) for _ in range(channel_num)]
            expected = p(
                np.stack([np.asarray((self.imt[i]) + factor[i]) for i in range(channel_num)]).astype(np.float32)
            )
            assert_allclose(result[key], p(expected), type_test="tensor")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import time
import unittest

import torch

from monai.data import DataLoader
from monai.utils import optional_import, set_determinism
from monai.utils.enums import CommonKeys
from tests.utils import SkipIfNoModule

try:
    _, has_ignite = optional_import("ignite")

    from monai.engines import SupervisedTrainer
    from monai.handlers import MetricLogger
    from monai.utils import ThreadContainer
except ImportError:
    has_ignite = False

compare_images, _ = optional_import("matplotlib.testing.compare", name="compare_images")


class TestThreadContainer(unittest.TestCase):
    @SkipIfNoModule("ignite")
    def test_container(self):
        net = torch.nn.Conv2d(1, 1, 3, padding=1)

        opt = torch.optim.Adam(net.parameters())

        img = torch.rand(1, 16, 16)
        data = {CommonKeys.IMAGE: img, CommonKeys.LABEL: img}
        loader = DataLoader([data for _ in range(10)])

        trainer = SupervisedTrainer(
            device=torch.device("cpu"),
            max_epochs=1,
            train_data_loader=loader,
            network=net,
            optimizer=opt,
            loss_function=torch.nn.L1Loss(),
        )

        con = ThreadContainer(trainer)
        con.start()
        time.sleep(1)  # wait for trainer to start

        self.assertTrue(con.is_alive)
        self.assertIsNotNone(con.status())
        self.assertTrue(len(con.status_dict) > 0)

        con.join()

    @SkipIfNoModule("ignite")
    @SkipIfNoModule("matplotlib")
    def test_plot(self):
        set_determinism(0)

        testing_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "testing_data")

        net = torch.nn.Conv2d(1, 1, 3, padding=1)

        opt = torch.optim.Adam(net.parameters())

        img = torch.rand(1, 16, 16)

        # a third non-image key is added to test that this is correctly ignored when plotting
        data = {CommonKeys.IMAGE: img, CommonKeys.LABEL: img, "Not Image Data": ["This isn't an image"]}

        loader = DataLoader([data] * 20, batch_size=2)

        trainer = SupervisedTrainer(
            device=torch.device("cpu"),
            max_epochs=1,
            train_data_loader=loader,
            network=net,
            optimizer=opt,
            loss_function=torch.nn.L1Loss(),
        )

        logger = MetricLogger()
        logger.attach(trainer)

        con = ThreadContainer(trainer)
        con.start()
        con.join()

        fig = con.plot_status(logger)

        with tempfile.TemporaryDirectory() as tempdir:
            tempimg = f"{tempdir}/threadcontainer_plot_test.png"
            fig.savefig(tempimg)
            comp = compare_images(f"{testing_dir}/threadcontainer_plot_test.png", tempimg, 5e-2)

            self.assertIsNone(comp, comp)  # None indicates test passed


if __name__ == "__main__":
    unittest.main()
